"""Unit tests for observation.py using function-style pytest tests."""

from collections import deque
from unittest.mock import patch

import pytest

from egse.observation import ObservationContext
from egse.observation import building_block
from egse.observation import end_observation
from egse.observation import execute
from egse.observation import request_obsid
from egse.observation import start_observation
from egse.observation import stringify_args
from egse.observation import stringify_kwargs
from egse.response import Failure
from egse.response import Success


@pytest.fixture
def obs_context():
    ctx = ObservationContext()
    ctx.level = -1
    ctx.bbid_count = 0
    ctx.bbids = deque()
    return ctx


def test_set_bbid_requires_observation_context(obs_context):
    with pytest.raises(RuntimeError, match="outside the scope of an observation"):
        obs_context.set_bbid("BBID-TEST")


def test_set_bbid_rejects_duplicates(obs_context):
    obs_context.level = 0
    obs_context.set_bbid("BBID-TEST")

    with pytest.raises(ValueError, match="already in execution"):
        obs_context.set_bbid("BBID-TEST")


def test_set_and_unset_bbid_updates_state(obs_context):
    obs_context.level = 0

    obs_context.set_bbid("BBID-TEST")
    assert obs_context.get_current_bbid() == "BBID-TEST"
    assert obs_context.get_level() == 1
    assert obs_context.bbid_count == 1

    obs_context.unset_bbid()
    assert obs_context.get_level() == 0
    assert list(obs_context.bbids) == []


def test_get_current_bbid_returns_latest(obs_context):
    obs_context.level = 0
    obs_context.set_bbid("BBID-1")
    obs_context.set_bbid("BBID-2")

    assert obs_context.get_current_bbid() == "BBID-2"


@patch("egse.observation.ConfigurationManagerProxy")
def test_start_observation_success(mock_cm, obs_context):
    mock_cm.return_value.__enter__.return_value.start_observation.return_value = Success("ok", 123)

    rc = obs_context.start_observation({"description": "test"})

    assert rc == 123
    assert obs_context.get_level() == 0


def test_observation_context_init_does_not_reset_shared_state(obs_context):
    obs_context.level = 5
    obs_context.bbid_count = 12
    obs_context.bbids.extend(["BBID-1"])

    another = ObservationContext()

    assert another.get_level() == 5
    assert another.bbid_count == 12
    assert list(another.bbids) == ["BBID-1"]


@patch("egse.observation.ConfigurationManagerProxy")
def test_start_observation_resets_stale_context_before_start(mock_cm, obs_context):
    obs_context.level = -1
    obs_context.bbid_count = 3
    obs_context.bbids.extend(["BBID-OLD"])

    mock_cm.return_value.__enter__.return_value.start_observation.return_value = Success("ok", 777)

    rc = obs_context.start_observation({"description": "test"})

    assert rc == 777
    assert obs_context.get_level() == 0
    assert obs_context.bbid_count == 0
    assert list(obs_context.bbids) == []


def test_start_observation_rejects_when_active(obs_context):
    obs_context.level = 0

    with pytest.raises(Failure, match="Only one Observation can be active"):
        obs_context.start_observation({"description": "test"})


@patch("egse.observation.ConfigurationManagerProxy")
def test_start_observation_failure_resets_level(mock_cm, obs_context):
    mock_cm.return_value.__enter__.return_value.start_observation.return_value = Failure("nope")

    with pytest.raises(Failure, match="nope"):
        obs_context.start_observation({"description": "test"})

    assert obs_context.get_level() == -1


@patch("egse.observation.ConfigurationManagerProxy")
def test_start_observation_connection_error(mock_cm, obs_context):
    mock_cm.return_value.__enter__.return_value.start_observation.side_effect = ConnectionError("boom")

    with pytest.raises(Failure, match="Couldn't connect"):
        obs_context.start_observation({"description": "test"})

    assert obs_context.get_level() == -1


@patch("egse.observation.ConfigurationManagerProxy")
def test_end_observation_success_resets_state(mock_cm, obs_context):
    obs_context.level = 1
    obs_context.bbid_count = 2
    obs_context.bbids.extend(["BBID-1", "BBID-2"])

    mock_cm.return_value.__enter__.return_value.end_observation.return_value = Success("ok")

    obs_context.end_observation()

    assert obs_context.get_level() == -1
    assert obs_context.bbid_count == 0
    assert list(obs_context.bbids) == []


@patch("egse.observation.ConfigurationManagerProxy")
def test_end_observation_failure_raises(mock_cm, obs_context):
    mock_cm.return_value.__enter__.return_value.end_observation.return_value = Failure("fail")

    with pytest.raises(Failure, match="fail"):
        obs_context.end_observation()


@patch("egse.observation.ConfigurationManagerProxy")
def test_request_obsid_returns_value(mock_cm):
    mock_cm.return_value.__enter__.return_value.get_obsid.return_value = Success("ok", 456)

    assert request_obsid() == 456


@patch("egse.observation.ConfigurationManagerProxy")
def test_request_obsid_raises_failure(mock_cm):
    mock_cm.return_value.__enter__.return_value.get_obsid.return_value = Failure("bad")

    with pytest.raises(Failure, match="bad"):
        request_obsid()


def test_stringify_args_handles_objects():
    class Sample:
        pass

    obj = Sample()
    args = stringify_args([obj, 42])

    assert args[0].endswith(".Sample")
    assert args[1] == "42"


def test_stringify_kwargs_handles_objects():
    class Sample:
        pass

    obj = Sample()
    kwargs = stringify_kwargs({"a": obj, "b": 7})

    assert kwargs["a"].endswith(".Sample")
    assert kwargs["b"] == "7"


@patch("egse.observation.request_obsid", return_value=111)
@patch("egse.observation.ConfigurationManagerProxy")
@patch("egse.observation.ObservationContext")
def test_execute_runs_building_block(mock_obs_ctx, mock_cm, mock_obsid):
    mock_cm.return_value.__enter__.return_value.start_observation.return_value = Success("ok", 123)
    mock_obs_ctx.level = 1

    @building_block
    def bb_func(x: int = 0) -> int:
        return x * 2

    result = execute(bb_func, "desc", x=3)
    assert result == 6


@patch("egse.observation.ObservationContext")
def test_execute_runs_building_block_missing_kwargs(mock_obs_ctx):
    with pytest.raises(ValueError, match="Expected 2 keyword parameters"):

        @building_block
        def bb(x: int = 1, y: int = 2) -> int:
            return x * y

        # This should fail since all kwargs are mandatory for the building block
        _ = execute(bb, "desc")

    assert not ObservationContext().is_active()  # Ensure context is not active after failure


@patch("egse.observation.request_obsid", return_value=111)
@patch("egse.observation.ObservationContext.end_observation")
@patch("egse.observation.ObservationContext.start_observation")
@patch("egse.observation.logger.warning")
def test_execute_runs_function_and_warns_when_not_building_block(
    mock_warn,
    mock_start,
    mock_end,
    mock_obsid,
):
    def plain_func(x):
        return x + 1

    result = execute(plain_func, "desc", 1)

    assert result == 2
    mock_warn.assert_called_once()
    mock_end.assert_called_once()


@patch("egse.observation.request_obsid", return_value=111)
@patch("egse.observation.ObservationContext.end_observation")
@patch("egse.observation.ObservationContext.start_observation")
def test_execute_calls_end_observation_on_exception(mock_start, mock_end, mock_obsid):
    def failing_func():
        raise RuntimeError("boom")

    with pytest.raises(RuntimeError, match="boom"):
        execute(failing_func, "desc")

    mock_end.assert_called_once()

    assert not ObservationContext().is_active()  # Ensure context is not active after exception


@patch("egse.observation.request_obsid", return_value=999)
@patch("egse.observation.ObservationContext.start_observation")
def test_start_observation_wrapper_returns_obsid(mock_start, mock_obsid):
    assert start_observation("testing") == 999


@patch("egse.observation.logger.error")
@patch("egse.observation.ObservationContext.start_observation")
def test_start_observation_wrapper_logs_and_raises(mock_start, mock_error):
    mock_start.side_effect = Failure("bad")

    with pytest.raises(Failure, match="bad"):
        start_observation("testing")

    mock_error.assert_called_once()


@patch("egse.observation.ObservationContext.end_observation")
def test_end_observation_wrapper_calls_context(mock_end):
    end_observation()
    mock_end.assert_called_once()



from egse.plugin import HierarchicalEntryPoints


def test_hierarchical_entry_points():
    print()

    cgse_ext = HierarchicalEntryPoints("cgse.extension")

    assert cgse_ext.base_group == "cgse.extension"
    assert len(cgse_ext.get_by_subgroup("setup_provider")) == 1



import logging

import numpy as np
import pytest
import transforms3d as t3
from pytest import approx

from egse.coordinates.reference_frame import ReferenceFrame
from egse.coordinates.rotation_matrix import RotationMatrix
from egse.exceptions import InvalidOperationError

LOGGER = logging.getLogger(__name__)


def test_master_construction():
    # This frame "refers to", i.e. it "is defined in", itself

    master = ReferenceFrame.create_master()

    assert np.array_equal(master.transformation, np.identity(4))
    assert master.name == "Master"
    assert master.reference_frame is master
    assert master.reference_frame == master
    assert master.rotation_config == "sxyz"


def test_link_to_master():
    master = ReferenceFrame.create_master()
    rot_config = "sxyz"

    glfix = ReferenceFrame(
        transformation=np.identity(4), reference_frame=master, name="glfix", rotation_config=rot_config
    )
    glfix.add_link(master, transformation=np.identity(4))

    assert not glfix.is_master()
    assert glfix.linked_to
    assert master.linked_to


def test_invalid_constructions():
    # We used to create a master frame when ref is None, but that is no longer supported.

    with pytest.raises(ValueError) as ve:
        ReferenceFrame(transformation=None, reference_frame=None)
    assert ve.value.args[1] == "REF_IS_NONE"

    # ref shall be a reference frame object

    with pytest.raises(ValueError) as ve:
        ReferenceFrame(transformation=None, reference_frame="MyReference")
    assert ve.value.args[1] == "REF_IS_NOT_CLS"

    # Master is a reserved name for the MASTER reference frame

    with pytest.raises(ValueError) as ve:
        master = ReferenceFrame.create_master()
        ReferenceFrame(transformation=None, reference_frame=master, name="Master")
    assert ve.value.args[1] == "MASTER_NAME_USED"

    # Master is a reserved name for the MASTER reference frame

    with pytest.raises(ValueError) as ve:
        master = ReferenceFrame.create_master()
        ReferenceFrame(transformation=[], reference_frame=master)
    assert ve.value.args[1] == "TRANSFORMATION_IS_NOT_NDARRAY"


def test_str():
    master = ReferenceFrame.create_master()

    assert str(master)


def test_repr():
    master = ReferenceFrame.create_master()

    assert repr(master) != str(master)

    # We want repr() to always return just one line

    assert "\n" not in repr(master)


def test_hash():
    # Any hash function MUST satisfy the following properties:
    #
    # - If two object are equal, then their hashes should be equal, i.e.
    #   a == b implies hash(a) == hash(b)
    #
    #   note: hash(a) == hash(b) does NOT imply a == b (which is a hash collision)
    #
    # - In order for an object to be hashable, it must be 'immutable', i.e. the
    #   hash of an object does not change across the object's lifetime
    #
    # For good hash functions, the following properties should be implemented:
    #
    # - If two objects have the same hash, then they are likely to be the same object
    #
    # - The hash of an object should be cheap to compute
    #
    # Check out this great article:
    #
    # What happens when you mess with hashing in Python [https://www.asmeurer.com/blog/posts/what-happens-when-you-mess-with-hashing-in-python/]

    master = ReferenceFrame.create_master()

    # In the current implementation of __eq__ all these reference frames will be
    # different because their name is different (and we cannot create two reference
    # frames with the same name).

    a1 = ReferenceFrame.from_translation(1, 1, 1, reference_frame=master, name="A1")
    a2 = ReferenceFrame.from_translation(1, 1, 1, reference_frame=master, name="A2")

    b1 = ReferenceFrame.from_translation(2, 0, 0, reference_frame=a1, name="B1")
    b2 = ReferenceFrame.from_translation(2, 0, 0, reference_frame=a2, name="B2")

    # Put the ReferenceFrames in a set (muttable), uses hashes
    frames = {master, a1, b1}  # __hash__ called three times
    assert master in frames  # __hash__ called once

    # Put the ReferenceFrames in a dict (muttable), uses hashes
    frames = {master: master, a1: a1, b1: b1}  # __hash__ called three times
    assert master in frames  # __hash__ called once

    assert master == master
    assert hash(master) == hash(master)


def test_add_link():
    master = ReferenceFrame.create_master()
    A = ReferenceFrame.from_translation(1, 1, 1, reference_frame=master, name="A")
    B = ReferenceFrame.from_translation(2, 0, 0, reference_frame=A, name="B")

    B.add_link(A, transformation=B.transformation)

    assert A in B.linked_to
    assert B in A.linked_to
    assert master not in A.linked_to
    assert master not in B.linked_to


def test_random_name():
    # The Master should have "Master" as its name
    master = ReferenceFrame.create_master()
    assert not master.name.startswith("F")

    # Any other reference frame that is not given a name should start with 'F'
    ref = ReferenceFrame.from_translation(1.0, 2.0, 3.0, reference_frame=master)
    assert ref.name.startswith("F")


def test_set_name():
    master = ReferenceFrame.create_master()
    with pytest.raises(InvalidOperationError):
        master.set_name("MyMaster")

    ref = ReferenceFrame.from_translation(1.0, 2.0, 3.0, reference_frame=master)
    assert ref.name != "Basic Translation"
    ref.set_name("Basic Translation")
    assert ref.name == "Basic Translation"


def test_translation():
    master = ReferenceFrame.create_master()

    # define a reference frame that is translated by 2 in Y

    transx, transy, transz = 0, 2, 0
    adef = np.identity(4)
    adef[:3, 3] = [transx, transy, transz]
    a = ReferenceFrame(transformation=adef, reference_frame=master, name="A")
    assert a is not None

    b = ReferenceFrame.from_translation(transx, transy, transz, master, name="B")
    assert b is not None

    assert np.array_equal(a.get_translation_vector(), b.get_translation_vector())
    assert a.reference_frame is b.reference_frame
    assert a is not b
    assert a != b
    assert a.is_same(b)


def test_rotation():
    master = ReferenceFrame.create_master()

    # Convention (rotating axes, in order xyz)

    rot_config = "rxyz"

    # Rotation amplitude

    rotx, roty, rotz = 0, 0, np.pi / 4.0

    rotation = RotationMatrix(rotx, roty, rotz, rot_config, active=True)

    # Defaults for zoom & shear

    zoom = np.array([1, 1, 1])
    shear = np.array([0, 0, 0])

    translation = [0, 0, 0]
    TT = t3.affines.compose(T=translation, R=rotation.rotation_matrix, Z=zoom, S=shear)

    # D is rotated wrt master

    D = ReferenceFrame(transformation=TT, reference_frame=master, name="D")

    E = ReferenceFrame.from_rotation(rotx, roty, rotz, master, rotation_config=rot_config, name="E", degrees=False)

    assert np.array_equal(D.get_rotation_matrix(), E.get_rotation_matrix())

    F = ReferenceFrame.from_rotation(rotx, roty, 45.0, reference_frame=master)

    assert np.array_equal(D.get_rotation_matrix(), F.get_rotation_matrix())
    assert F.is_same(D)


def test_equals():
    master = ReferenceFrame.create_master()

    assert master is master
    assert master == master

    m1 = ReferenceFrame.create_master()
    m2 = ReferenceFrame.create_master()

    assert m1 is not master
    assert m1 == master
    assert m1 is not m2
    assert m1 == m2

    t1 = ReferenceFrame.from_translation(1, 2, 3, reference_frame=master)
    t2 = ReferenceFrame.from_translation(1, 2, 3, reference_frame=master)
    t3 = ReferenceFrame.from_translation(1, 2, 3, reference_frame=master, name=t2.name)
    t4 = ReferenceFrame.from_translation(2, 3, 4, reference_frame=master)
    t5 = ReferenceFrame.from_translation(2, 3, 4, reference_frame=m1)
    t6 = ReferenceFrame.from_translation(2, 3, 4, reference_frame=t2)

    assert t1 != t2
    assert t1 is not t2
    assert t1.is_same(t2)
    assert t2.is_same(t1)
    assert t1 != t3
    # t3 will be given another random generated name as t2.name exists already,
    # i.e. no two reference frames can have the same name
    # !! This rule has been relaxed and we now allow two or even more ReferenceFrames with the
    # !! same name, therefore != changed into ==
    assert t2 == t3
    assert t2.name == t3.name
    assert t2.is_same(t3)
    assert not t3.is_same(t4)
    assert t5 != t4
    assert t4.is_same(t5)
    assert t6 != t4  # different ref
    assert not t5.is_same(t6)

    r1 = ReferenceFrame.from_rotation(1, 2, 3, reference_frame=master)
    r2 = ReferenceFrame.from_rotation(1, 2, 3, reference_frame=master)
    r3 = ReferenceFrame.from_rotation(1, 2, 3, reference_frame=master, name=r2.name)
    r4 = ReferenceFrame.from_rotation(2, 3, 4, reference_frame=master)
    r5 = ReferenceFrame.from_rotation(2, 3, 4, reference_frame=m2)
    r6 = ReferenceFrame.from_rotation(2, 3, 4, reference_frame=r2)

    assert r1 != r2
    assert r1 is not r2
    assert r1.is_same(r2)
    assert r2.is_same(r1)
    assert r1 != r3
    # t3 will be given another random generated name as t2.name exists already,
    # i.e. no two reference frames can have the same name
    # !! This rule has been relaxed and we now allow two or even more ReferenceFrames with the
    # !! same name, therefore != changed into ==
    assert r2 == r3
    assert r2.is_same(r3)
    assert r2.name == r3.name
    assert r4 != r5
    assert r4.is_same(r5)
    assert r5 != r6  # different ref
    assert not r5.is_same(r6)

    assert master != "any other object"


def test_copy():
    import copy

    master = ReferenceFrame.create_master()

    assert master is copy.copy(master)
    assert master == copy.copy(master)

    r = ReferenceFrame.from_translation(1, 2, 3, reference_frame=master)

    assert r is not copy.copy(r)

    # This next test changed from != into == since we have relaxed the rule on
    # unique naming of ReferenceFrames
    assert r == copy.copy(r)

    assert r.is_same(copy.copy(r))


def test_position_after_homing():
    # Rotation around static axis, and around x, y and z in that order

    rot_config = "sxyz"

    # Use degrees in all arguments

    degrees = True

    # Configure representations of the coordinate systems of the hexapod
    # Configure the following reference frames: master, mec, usr, plt, obj, obusr
    # Configure invariant links between those reference frames

    master = ReferenceFrame.create_master()

    # MEC = MASTER
    mec = ReferenceFrame(transformation=np.identity(4), reference_frame=master, name="mec", rotation_config=rot_config)

    # USR, defined in MEC
    tr_u = np.array([0, 0, 0])
    rot_u = np.array([0, 0, 0])

    usr = ReferenceFrame.from_translation_rotation(
        tr_u, rot_u, rotation_config=rot_config, reference_frame=mec, name="usr", degrees=degrees
    )

    # PLATFORM (default after homing: PLT = MEC)
    plt = ReferenceFrame(transformation=np.identity(4), reference_frame=mec, name="plt", rotation_config=rot_config)

    # OBJECT, defined wrt PLATFORM
    tr_o = np.array([0, 0, 0])
    rot_o = np.array([0, 0, 0])
    obj = ReferenceFrame.from_translation_rotation(
        tr_o, rot_o, rotation_config=rot_config, reference_frame=plt, name="obj", degrees=degrees
    )

    # OBUSR == OBJ, but defined wrt USR  (OBJ is defined in PLT) --> used in moveAbsolute
    transfo = usr.get_active_transformation_to(obj)
    obusr = ReferenceFrame(transfo, rotation_config=rot_config, reference_frame=usr, name="obusr")

    # Configure the invariant links within the system

    # Link OBUSR to OBJ
    # OBUSR = OBJ ==> the link is the identity matrix
    obusr.add_link(obj, transformation=np.identity(4))

    # Link PLT to OBJ
    # The link is the definition of OBJ given above

    transformation = obj.transformation
    plt.add_link(obj, transformation=transformation)

    # Link USR to MEC
    # The link is the definition of USR given above

    transformation = usr.transformation
    mec.add_link(usr, transformation=transformation)

    # Now perfom the actual tests.

    # Check system status before any movement, i.e. the default initialization state

    out = plt.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])
    out = obusr.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])
    out = usr.get_active_translation_rotation_vectors_to(obusr)
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])

    # Configure movement

    tx, ty, tz = [5, 2, 0]
    rx, ry, rz = [0, 0, 0]

    tr_abs = np.array([tx, ty, tz])
    rot_abs = np.array([rx, ry, rz])

    obusr.set_translation_rotation(
        tr_abs, rot_abs, rotation_config=rot_config, active=True, degrees=True, preserve_links=True
    )

    out = plt.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [5, 2, 0, 0, 0, 0])
    out = obusr.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [5, 2, 0, 0, 0, 0])
    out = usr.get_active_translation_rotation_vectors_to(obusr)
    check_positions(np.reshape(out, 6), [5, 2, 0, 0, 0, 0])

    # Perform a Homing, i.e. goto position zero

    tx, ty, tz = [0, 0, 0]
    rx, ry, rz = [0, 0, 0]

    tr_abs = np.array([tx, ty, tz])
    rot_abs = np.array([rx, ry, rz])

    # See issue #58
    # plt.setTranslationRotation(tr_abs,rot_abs,rot_config=rot_config, active=True, degrees=True,preserveLinks=True)

    tr_abs, rot_abs = obj.get_translation_rotation_vectors()
    obusr.set_translation_rotation(
        tr_abs, rot_abs, rotation_config=rot_config, active=True, degrees=True, preserve_links=True
    )

    out = plt.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])
    out = obusr.get_translation_rotation_vectors()
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])
    out = usr.get_active_translation_rotation_vectors_to(obusr)
    check_positions(np.reshape(out, 6), [0, 0, 0, 0, 0, 0])


def test_linked_to_reference():
    master = ReferenceFrame.create_master()

    translation = [0, 2, 0]
    Adef = np.identity(4)
    Adef[:3, 3] = translation
    A1 = ReferenceFrame(transformation=Adef, reference_frame=master, name="A1")
    A2 = ReferenceFrame(transformation=Adef, reference_frame=master, name="A2")

    # C
    translation = [2, 0, 0]
    Cdef = np.identity(4)
    Cdef[:3, 3] = translation
    C = ReferenceFrame(transformation=Cdef, reference_frame=master, name="C")

    # B
    translation = [2, 0, 0]
    Bdef = np.identity(4)
    Bdef[:3, 3] = translation
    B1 = ReferenceFrame(transformation=Bdef, reference_frame=A1, name="B1")
    B2 = ReferenceFrame(transformation=Bdef, reference_frame=A2, name="B2")

    # Frame D is defined in C
    translation = [2, 0, 0]
    rotation = [0, 0, 45]
    degrees = True
    D = ReferenceFrame.from_translation_rotation(
        translation=translation, rotation=rotation, reference_frame=C, name="D", degrees=degrees
    )

    B1.add_link(A1, transformation=B1.transformation)

    translation = [0, 0, 0]
    rotation = [0, 0, 45]
    A1.apply_translation_rotation(translation, rotation, active=True, degrees=True)
    A2.apply_translation_rotation(translation, rotation, active=True, degrees=True)

    assert A1.is_same(A2)
    assert not B1.is_same(B2)


def check_positions(out, expected, precision=0.00001):
    assert len(out) == len(expected)

    for idx, element in enumerate(out):
        assert element == approx(expected[idx], precision)



import time
import uuid
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True, slots=True)
class NotificationEvent:
    event_type: str
    source_service: str
    data: dict
    timestamp: float | None = None
    correlation_id: str | None = None

    def __post_init__(self):
        if self.correlation_id is None:
            object.__setattr__(self, "correlation_id", str(uuid.uuid4()))

        if self.timestamp is None:
            object.__setattr__(self, "timestamp", time.time())

    def as_dict(self):
        return {
            "event_type": self.event_type,
            "source_service": self.source_service,
            "data": self.data,
            "timestamp": self.timestamp,
            "correlation_id": self.correlation_id,
        }



"""
This module defines the level, format and handlers for the root logger and for the special
'egse' logger. The egse_logger will be configured with a special handler which sends all
logging messages to a log control server.

This module is loaded whenever an egse module is loaded, to ensure all log messages are properly
forwarded to the log control server.
"""

__all__ = [
    "SERVICE_TYPE",
    "LOG_FORMAT_FULL",
    "logger",
    "root_logger",
    "egse_logger",
    "close_all_zmq_handlers",
    "create_new_zmq_logger",
    "get_log_file_name",
    "print_all_handlers",
    "replace_zmq_handler",
    "send_request",
    "set_all_logger_levels",
    "setup_logging",
    "teardown_logging",
    "remote_logging",
]

import contextlib
import logging
import pickle
import traceback

import zmq

from egse.log import LOG_FORMAT_FULL
from egse.log import PackageFilter
from egse.log import egse_logger
from egse.log import get_log_level_from_env
from egse.log import logger
from egse.log import root_logger
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.system import is_in_ipython
from egse.system import type_name

settings = Settings.load("Logging Control Server")

SERVICE_TYPE = settings.get("SERVICE_TYPE", "log_cs")
"""The logger id that is also used as service_type for service discovery."""

PROTOCOL = settings.get("PROTOCOL", "tcp")
HOSTNAME = settings.get("HOSTNAME", "localhost")
RECEIVER_PORT = settings.get("RECEIVER_PORT", 0)  # dynamically assigned by the system if 0
COMMANDER_PORT = settings.get("COMMANDER_PORT", 0)  # dynamically assigned by the system if 0


_initialised = False  # will be set to True in the setup_logging() function


def get_log_file_name() -> str:
    """
    Returns the filename of the log file as defined in the Settings or return the default name 'general.log'.
    """
    return settings.get("FILENAME", "general.log")


def get_endpoint_from_registry(service_type: str, port_type: str = "commander"):
    """Returns the endpoint that was constructed from information from the service registry."""

    # logger.debug(f"Calling get_endpoint_from_registry(\"{port}\")")

    try:
        with RegistryClient() as reg:
            if port_type == "commander":
                endpoint = reg.get_endpoint(service_type)
            elif port_type == "receiver":
                service = reg.discover_service(service_type)
                endpoint = (
                    f"{service.get('protocol', 'tcp')}://{service['host']}:{service['metadata']['receiver_port']}"
                )
            else:
                logger.error(f"Incorrect argument {port_type=}, no endpoint returned.")
                endpoint = None
    except Exception:  # noqa
        logger.warning(f"Couldn't retrieve endpoint for {service_type}. Is the service registry process running?")
        endpoint = None

    return endpoint


class ZeroMQHandler(logging.Handler):
    def __init__(self):
        super().__init__()

        if RECEIVER_PORT == 0:
            endpoint = get_endpoint_from_registry(service_type=SERVICE_TYPE, port_type="receiver")
        else:
            endpoint = f"{PROTOCOL}://{HOSTNAME}:{RECEIVER_PORT}"

        if endpoint:
            self.setLevel(logging.NOTSET)

            self.ctx = zmq.Context.instance()
            self.socket = zmq.Socket(self.ctx, zmq.PUSH)
            self.socket.setsockopt(zmq.SNDHWM, 0)  # never block on sending msg
            # self.socket.setsockopt(zmq.RCVTIMEO, 1000)  # 1 second
            # self.socket.setsockopt(zmq.SNDTIMEO, 1000)  # 1 second
            self.socket.setsockopt(zmq.LINGER, 100)  # short wait for unsent messages on close()
            self.socket.setsockopt(zmq.IMMEDIATE, 1)  # set immediate mode
            self.socket.connect(endpoint)

            logger.debug(f"All logging messages will be forwarded to the LOGGER at {endpoint}.")
        else:
            self.ctx = None
            self.socket = None

            logger.warning("No logging messages will be forwarded to the LOGGER")

    def __del__(self):
        self.close()

    def close(self):
        if self.socket:
            self.socket.close(linger=100)
            self.socket = None

    def emit(self, record):
        """
        Emit a record.

        Writes the LogRecord to the queue, preparing it for pickling first.
        """

        if not self.socket:
            return

        try:
            if record.exc_info:
                record.exc_text = traceback.format_exc()
                record.exc_info = None  # traceback objects can not be pickled
            if record.processName == "MainProcess" and is_in_ipython():
                record.processName = "IPython"
            data = pickle.dumps(record.__dict__)
            self.socket.send(data, flags=zmq.NOBLOCK)
        except (KeyboardInterrupt, SystemExit):
            raise
        except zmq.Again:
            # This exception occurs when e.g. the log_cs server has not yet started
            # but logging was set up in another process. An example is the service
            # registry which starts before the logger. When it calls `setup_logging()`
            # the `log_cs` has not yet started.
            pass
        except Exception as exc:
            logging.error(f"ZeroMQHandler: Caught {type_name(exc)} - {exc}", exc_info=True)
            self.handleError(record)


def print_all_handlers():
    """A debugging function to check which handlers are connected to loggers."""

    print("=" * 80)
    print("Root logger handlers:")
    for i, handler in enumerate(logging.root.handlers):
        print(f"  {i}: {type(handler).__name__} - {handler}")

    print("\nAll loggers with handlers:")
    for name, logger in logging.Logger.manager.loggerDict.items():
        if hasattr(logger, "handlers") and logger.handlers:
            print(f"  {name}: {[type(h).__name__ for h in logger.handlers]}")

    print("\n" + "=" * 80)


def close_all_zmq_handlers():
    """
    Close and remove all the ZeroMQHandlers that are connected to a logger.
    """

    loggers = logging.Logger.manager.loggerDict

    for name, logger in loggers.items():
        if isinstance(logger, logging.PlaceHolder):
            continue
        for handler in logger.handlers:
            if isinstance(handler, ZeroMQHandler):
                root_logger.debug(f"Close and remove ZeroMQHandler for logger {name}.")
                handler.close()
                logger.removeHandler(handler)


def setup_logging():
    # Initialize logging as we want it for the Common-EGSE
    #
    # * The ZeroMQHandler to send all logging messages, i.e. level=DEBUG to the Logging Server
    # * The (local) StreamingHandlers to print only INFO messages and higher

    global root_logger, egse_logger, _initialised

    if _initialised:
        logger.info("Logging has already been set up for this process.")
        return

    logging.disable(logging.NOTSET)
    for handler in root_logger.handlers:
        handler.setLevel(get_log_level_from_env())

    egse_logger.setLevel(logging.DEBUG)

    # Add the ZeroMQHandler to the egse_logger

    zmq_handler = ZeroMQHandler()
    zmq_handler.setLevel(logging.NOTSET)
    zmq_handler.addFilter(PackageFilter())
    egse_logger.addHandler(zmq_handler)

    _initialised = True


def teardown_logging():
    global _initialised

    close_all_zmq_handlers()

    _initialised = False


@contextlib.contextmanager
def remote_logging():
    """A context manager that can be used to activate remote logging."""
    setup_logging()
    yield
    teardown_logging()


# NOTE: This code is commented, you will have to set up and tear down the logging for
#       each process individually.
# setup_logging()
# atexit.register(teardown_logging)


def replace_zmq_handler():
    """
    This function will replace the current ZeroMQ Handler with a new instance. Use this function
    in the run() method of a multiprocessing.Process:

        >>> import egse.logger
        >>> egse.logger.replace_zmq_handler()

    Don't use this function in the __init__() method as only the run() method will execute in
    the new Process and replace the handler in the proper environment. The reason for this is
    that the ZeroMQ socket is not thread/Process safe, so a new ZeroMQ socket needs to be created
    in the correct process environment.
    """
    global egse_logger

    this_handler = None
    for handler in egse_logger.handlers:
        if isinstance(handler, ZeroMQHandler):
            this_handler = handler
    if this_handler is not None:
        egse_logger.removeHandler(this_handler)
    egse_logger.addHandler(ZeroMQHandler())


def create_new_zmq_logger(name: str) -> logging.Logger:
    """
    Create a new logger with the given name and add a ZeroMQ Handler to this logger.

    If the logger already has a ZeroMQ handler attached, don't add a second ZeroMQ handler,
    just return the Logger object.

    Args:
        name: the requested name for the logger

    Returns:
        A Logger for the given name with a ZeroMQ handler attached.
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # If the ZeroMQ handler already exists for this logger, don't add a second handler

    for handler in logger.handlers:
        if isinstance(handler, ZeroMQHandler):
            return logger

    zmq_handler = ZeroMQHandler()
    zmq_handler.setLevel(logging.NOTSET)

    logger.addHandler(zmq_handler)
    logger.setLevel(logging.DEBUG)

    return logger


def set_all_logger_levels(level: int):
    global root_logger, egse_logger

    root_logger.level = level
    egse_logger.level = level

    for handler in root_logger.handlers:
        handler.setLevel(level)

    # We don't want to restrict egse_logger levels

    # for handler in egse_logger.handlers:
    #     handler.setLevel(level)


TIMEOUT_RECV = 1.0  # seconds


def send_request(command_request: str):
    """Sends a request to the Logger Control Server and waits for a response."""

    if COMMANDER_PORT == 0:
        endpoint = get_endpoint_from_registry(SERVICE_TYPE)
    else:
        endpoint = f"{PROTOCOL}://{HOSTNAME}:{COMMANDER_PORT}"

    if endpoint is None:
        return {"error": "Couldn't determine the endpoint for the logger."}

    ctx = zmq.Context.instance()

    socket = ctx.socket(zmq.REQ)
    socket.connect(endpoint)

    socket.send(pickle.dumps(command_request))
    rlist, _, _ = zmq.select([socket], [], [], timeout=TIMEOUT_RECV)
    if socket in rlist:
        response = socket.recv()
        response = pickle.loads(response)
    else:
        response = {"error": "The ZeroMQ socket timed out for the Logger Control Server."}
    socket.close(linger=0)

    return response



from pathlib import Path

from egse.settings import Settings

HERE = Path(__file__).parent
settings = Settings.load("Digilent MEASURpoint DT8874 Control Server")

# General information about the Digilent MEASURpoint DT8874 Control Server

PROCESS_NAME = settings.get("PROCESS_NAME", "dt8874_cs")  # Name under which it is registered in the service registry
SERVICE_TYPE = settings.get(
    "SERVICE_TYPE", "dt8874_cs"
)  # Service type under which it is registered in the service registry
PROTOCOL = settings.get("PROTOCOL", "tcp")  # Communication protocol
HOSTNAME = settings.get("HOSTNAME", "localhost")  # Hostname
COMMANDING_PORT = settings.get("COMMANDING_PORT", 0)  # Commanding port (as per settings or dynamically assigned)
SERVICE_PORT = settings.get("SERVICE_PORT", 0)  # Service port (as per settings or dynamically assigned)
MONITORING_PORT = settings.get("MONITORING_PORT", 0)  # Monitoring port (as per settings or dynamically assigned)
ORIGIN = settings.get("STORAGE_MNEMONIC", "DT8874")  # Storage mnemonic (used in the HK filenames)

PROXY_TIMEOUT = 10



import subprocess
from typing import Annotated

import sys

import rich
import typer

from egse.system import redirect_output_to_log, all_logging_disabled

dt8874 = typer.Typer(
    name="dt8874", help="Digilent MEASURpoint DT8874, temperature and voltage monitoring", no_args_is_help=True
)


@dt8874.command(name="start")
def start_dt8874(
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="use a device simulator as the backend")
    ] = False,
):
    """Start the Digilent MEASURpoint DT8874 Control Server.

    The Control Server is always started in the background.
    """

    rich.print(f"Starting the Digilent MEASURpoint DT8874 Control Server - {simulator = }")
    out = redirect_output_to_log("dt8874_cs.start.log")

    cmd = [sys.executable, "-m", "egse.digilent.measurpoint.dt8874.dt8874_cs", "start"]
    if simulator:
        cmd.append("--simulator")

    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


@dt8874.command(name="stop")
def stop_dt8874():
    """Stops the Digilent MEASURpoint DT8874 Control Server."""

    rich.print("Stopping the Digilent MEASURpoint DT8874 Control Server")
    out = redirect_output_to_log("dt8874_cs.stop.log")

    cmd = [sys.executable, "-m", "egse.digilent.measurpoint.dt8874.dt8874_cs", "stop"]
    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


@dt8874.command(name="status")
def status_dt8874():
    """Prints the status information about the Digilent MEASURpoint DT8874 Control Server."""

    with all_logging_disabled():
        from egse.digilent.measurpoint.dt8874 import dt8874_cs

        dt8874_cs.status()


if __name__ == "__main__":
    dt8874()



import asyncio
from typing import Any
from typing import cast

try:
    from typing import override
except ImportError:
    try:
        from typing_extensions import override
    except ImportError:

        def override(method, /):
            return method


import pytest
from fixtures.helpers import capture_log_records

from egse.async_control import AcquisitionAsyncControlServer
from egse.async_control import AsyncControlClient
from egse.async_control import AsyncControlServer
from egse.async_control import InitializationError
from egse.async_control import is_control_server_active
from egse.async_dummy import DummyAsyncControlClient
from egse.async_dummy import DummyAsyncControlServer
from egse.log import logging
from egse.logger import remote_logging
from egse.system import type_name

# pytestmark = pytest.mark.skip("Implementation and tests are still a WIP")

logger = logging.getLogger("egse.tests.async_control")


class RecordingAsyncControlServer(AcquisitionAsyncControlServer):
    """Test helper server that records acquisition flow for assertions."""

    def __init__(self):
        """Initialize in-memory buffers used by acquisition tests."""
        super().__init__()
        self.processed_records: list[dict[str, Any]] = []
        self.processed_batches: list[list[dict[str, Any]]] = []
        self.processed_event = asyncio.Event()
        self.expected_records = 0

    @override
    async def handle_acquisition_batch(self, records: list[dict[str, Any]]):
        """Capture batch boundaries before delegating to per-record processing."""
        self.processed_batches.append(records.copy())
        await super().handle_acquisition_batch(records)

    @override
    async def handle_acquisition_record(self, record: dict[str, Any]):
        """Capture records and signal when the expected test count is reached."""
        self.processed_records.append(record)
        if len(self.processed_records) >= self.expected_records:
            self.processed_event.set()


class SequentialExecutionTestServer(AsyncControlServer):
    """Test helper server exposing sequential execution behavior."""

    def __init__(self):
        super().__init__()
        self.execution_order: list[int] = []

    async def append_after(self, token: int, delay: float):
        await asyncio.sleep(delay)
        self.execution_order.append(token)
        return token


@pytest.mark.asyncio
async def test_control_server(caplog):
    # First start the control server as a background task.
    server = AsyncControlServer()
    server_task = asyncio.create_task(server.start())

    await asyncio.sleep(0.5)  # give the server time to startup

    try:
        # Now create a control client that will connect to the above server.
        async with AsyncControlClient(service_type="async-control-server") as client:
            caplog.clear()

            # Sleep some time, so we can see the control server in action, e.g. status reports, housekeeping, etc
            await asyncio.sleep(5.0)

            assert "Sending status updates" in caplog.text  # this should there be 5 times actually

            response = await client.ping()
            logger.debug(f"{response = }")
            assert isinstance(response, str)
            assert response == "pong"

            response = await client.info()
            logger.debug(f"{response = }")
            assert isinstance(response, dict)
            assert "name" in response
            assert "hostname" in response
            assert "device commanding port" in response
            assert "service commanding port" in response

            assert await is_control_server_active(service_type="async-control-server")

            response = await client.stop_server()
            logger.debug(f"{response = }")
            assert isinstance(response, dict)
            assert response["status"] == "terminating"

            assert await is_control_server_active(service_type="async-control-server")
    except InitializationError as exc:
        logger.error(f"Could not create AsyncControlClient: {type_name(exc)}: {exc}")

    server_task.cancel()
    await server_task

    assert not await is_control_server_active(service_type="async-control-server")


@pytest.mark.asyncio
async def test_dummy_control_server():
    """Test DummyAsyncControlServer as an implementation example."""

    with remote_logging():
        server = DummyAsyncControlServer()
        server_task = asyncio.create_task(server.start())

        await asyncio.sleep(0.5)  # give the server time to startup

        # Now create a control client that will connect to the above server.
        async with DummyAsyncControlClient() as client:
            response = await client.ping()
            logger.debug(f"{response = }")
            assert isinstance(response, str)
            assert response == "pong"

            response = await client.echo("dummy")
            logger.debug(f"{response = }")
            assert response == "dummy"

            response = await client.set_value("configured")
            logger.debug(f"{response = }")
            assert response == "configured"

            info_response = await client.info()
            logger.debug(f"{info_response = }")
            assert isinstance(info_response, dict)
            info_response = cast(dict[str, object], info_response)
            assert "name" in info_response
            assert "hostname" in info_response
            assert "device commanding port" in info_response
            assert "service commanding port" in info_response
            assert info_response["service type"] == DummyAsyncControlServer.service_type
            assert info_response["echo count"] == 1
            assert info_response["last value"] == "configured"

            health_response = await client.health()
            logger.debug(f"{health_response = }")
            assert isinstance(health_response, dict)
            assert health_response["status"] == "ok"
            assert health_response["echo count"] == 1
            assert health_response["last value"] == "configured"

            start_response = await client.start_acquisition(interval=0.02)
            logger.debug(f"{start_response = }")
            assert isinstance(start_response, dict)
            assert start_response["running"] is True

            await asyncio.sleep(1.5)  # this should allow for ~75 records to be acquired at 20ms intervals

            stop_response = await client.stop_acquisition()
            logger.debug(f"{stop_response = }")
            assert isinstance(stop_response, dict)
            assert stop_response["running"] is False
            assert int(stop_response["acquisition logged"]) > 0

            health_response = await client.health()
            logger.debug(f"{health_response = }")
            assert isinstance(health_response, dict)
            assert health_response["acquisition running"] is False
            assert int(health_response["acquisition logged"]) > 0

        server_task.cancel()
        await server_task


@pytest.mark.asyncio
async def test_acquisition_callback_from_thread_is_processed_in_order():
    server = RecordingAsyncControlServer()
    server.expected_records = 3
    server._loop = asyncio.get_running_loop()

    processing_task = asyncio.create_task(server.process_acquisition_data())

    try:
        callback = cast(Any, server.get_acquisition_callback())

        await asyncio.to_thread(callback, {"value": 1}, source="sensor-a", metadata={"index": 1})
        await asyncio.to_thread(callback, {"value": 2}, source="sensor-a", metadata={"index": 2})
        await asyncio.to_thread(callback, {"value": 3}, source="sensor-a", metadata={"index": 3})

        await asyncio.wait_for(server.processed_event.wait(), timeout=1.0)

        assert [record["data"]["value"] for record in server.processed_records] == [1, 2, 3]
        assert [record["metadata"]["index"] for record in server.processed_records] == [1, 2, 3]
        assert all(record["source"] == "sensor-a" for record in server.processed_records)
        assert server.acquisition_dropped_count == 0
        assert len(server.processed_batches) == 3
        assert all(len(batch) == 1 for batch in server.processed_batches)
    finally:
        server.stop()
        processing_task.cancel()
        await asyncio.gather(processing_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_acquisition_callback_same_thread_defaults_to_single_record_processing():
    server = RecordingAsyncControlServer()
    server.expected_records = 3
    server._loop = asyncio.get_running_loop()

    processing_task = asyncio.create_task(server.process_acquisition_data())

    try:
        callback = cast(Any, server.get_acquisition_callback())

        callback("first", source="sensor-local", metadata={"index": 1})
        callback("second", source="sensor-local", metadata={"index": 2})
        callback("third", source="sensor-local", metadata={"index": 3})

        await asyncio.wait_for(server.processed_event.wait(), timeout=1.0)

        assert [record["data"] for record in server.processed_records] == ["first", "second", "third"]
        assert [record["metadata"]["index"] for record in server.processed_records] == [1, 2, 3]
        assert all(record["source"] == "sensor-local" for record in server.processed_records)
        assert server.acquisition_batch_enabled is False
        assert len(server.processed_batches) == 3
        assert all(len(batch) == 1 for batch in server.processed_batches)
    finally:
        server.stop()
        processing_task.cancel()
        await asyncio.gather(processing_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_acquisition_batching_is_optional_and_preserves_order():
    server = RecordingAsyncControlServer()
    server.expected_records = 3
    server.acquisition_batch_enabled = True
    server.acquisition_batch_max_size = 10
    server.acquisition_batch_max_wait_s = 0.2
    server._loop = asyncio.get_running_loop()

    processing_task = asyncio.create_task(server.process_acquisition_data())

    try:
        callback = cast(Any, server.get_acquisition_callback())

        callback("first", source="sensor-b", metadata={"index": 1})
        callback("second", source="sensor-b", metadata={"index": 2})
        callback("third", source="sensor-b", metadata={"index": 3})

        await asyncio.wait_for(server.processed_event.wait(), timeout=1.0)

        assert [record["data"] for record in server.processed_records] == ["first", "second", "third"]
        assert [record["metadata"]["index"] for record in server.processed_records] == [1, 2, 3]
        assert len(server.processed_batches) >= 1
        assert [record["data"] for record in server.processed_batches[0]] == ["first", "second", "third"]
    finally:
        server.stop()
        processing_task.cancel()
        await asyncio.gather(processing_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_acquisition_queue_overflow_increments_drop_count():
    server = RecordingAsyncControlServer()
    server._loop = asyncio.get_running_loop()
    server._acquisition_queue = asyncio.Queue(maxsize=2)

    callback = cast(Any, server.get_acquisition_callback())

    callback("first", source="sensor-overflow")
    callback("second", source="sensor-overflow")
    callback("third", source="sensor-overflow")

    await asyncio.sleep(0)

    assert server._acquisition_queue.qsize() == 2
    assert server.acquisition_dropped_count == 1


@pytest.mark.asyncio
async def test_execute_sequential_serializes_concurrent_submissions_in_enqueue_order():
    server = SequentialExecutionTestServer()
    queue_task = asyncio.create_task(server.process_sequential_queue())

    try:
        tasks = [
            asyncio.create_task(server._execute_sequential(server.append_after(1, 0.03))),
            asyncio.create_task(server._execute_sequential(server.append_after(2, 0.01))),
            asyncio.create_task(server._execute_sequential(server.append_after(3, 0.0))),
        ]

        results = await asyncio.gather(*tasks)

        assert results == [1, 2, 3]
        assert server.execution_order == [1, 2, 3]
    finally:
        server.stop()
        queue_task.cancel()
        await asyncio.gather(queue_task, return_exceptions=True)


@pytest.mark.asyncio
async def test_execute_sequential_propagates_operation_exceptions():
    server = SequentialExecutionTestServer()
    queue_task = asyncio.create_task(server.process_sequential_queue())

    async def failing_operation():
        raise RuntimeError("boom")

    try:
        with pytest.raises(RuntimeError, match="boom"):
            await server._execute_sequential(failing_operation())
    finally:
        server.stop()
        queue_task.cancel()
        await asyncio.gather(queue_task, return_exceptions=True)


if __name__ == "__main__":
    with capture_log_records("egse") as caplog:
        asyncio.run(test_control_server(caplog=caplog))



""" " This module contains plotting utility functions for reference frames."""

import matplotlib
import matplotlib.pyplot as plt
from egse.coordinates.reference_frame import ReferenceFrame
import numpy as np
import egse
from egse.coordinates.point import Point, Points


def plot_reference_frame(
    frame: ReferenceFrame, master: ReferenceFrame | None = None, fig_name: str | None = None, **kwargs
):
    """Plots a reference frame.

    Use ax.set_xlim3d(min, max) to properly set the ranges of the display.

    Args:
        frame (ReferenceFrame): Reference frame
        master (ReferenceFrame | None): Optional master reference frame.
        fig_name (str): Name of the plot figure.
        kwargs : Keyword arguments passed to matplotlib.axes._subplots.Axes3DSubplot.quiver.

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the reference frame.  The three unit vectors are shown in
        the following colours (RGB):
            x: Red
            y: Green
            z: Blue
    """

    if master is None:
        temp_master = ReferenceFrame.create_master()
    else:
        temp_master = master.__copy__()

    # Origin and unit axes of the reference frame

    origin = frame.get_origin()
    unit_axis_x = frame.get_axis("x", name="fx")
    unit_axis_y = frame.get_axis("y", name="fy")
    unit_axis_z = frame.get_axis("z", name="fz")

    # Now expressed in the master reference frame

    origin_master = origin.express_in(temp_master)[:3]
    unit_axis_x_master = unit_axis_x.express_in(temp_master)[:3]
    unit_axis_y_master = unit_axis_y.express_in(temp_master)[:3]
    z_axis_master = unit_axis_z.express_in(temp_master)[:3]

    del temp_master

    # Coordinates of the origin, expressed in the master reference frame
    origin_x_master, origin_y_master, origin_z_master = (
        np.array([origin_master[0]]),
        np.array([origin_master[1]]),
        np.array([origin_master[2]]),
    )

    # Orientation of the unit vectors of the reference frame, expressed in the master reference frame

    vecxx, vecyx, veczx = (
        np.array([unit_axis_x_master[0] - origin_master[0]]),
        np.array([unit_axis_y_master[0] - origin_master[0]]),
        np.array([z_axis_master[0] - origin_master[0]]),
    )
    vecxy, vecyy, veczy = (
        np.array([unit_axis_x_master[1] - origin_master[1]]),
        np.array([unit_axis_y_master[1] - origin_master[1]]),
        np.array([z_axis_master[1] - origin_master[1]]),
    )
    vecxz, vecyz, veczz = (
        np.array([unit_axis_x_master[2] - origin_master[2]]),
        np.array([unit_axis_y_master[2] - origin_master[2]]),
        np.array([z_axis_master[2] - origin_master[2]]),
    )

    kwargs.setdefault("length", 1)
    kwargs.setdefault("normalize", True)
    # kwargs.setdefault('figsize', (10,10))

    fig = plt.figure(fig_name, figsize=plt.figaspect(1.0))
    ax = fig.add_subplot(projection="3d")
    ax.quiver(origin_x_master, origin_y_master, origin_z_master, vecxx, vecxy, vecxz, color="r", **kwargs)
    ax.quiver(origin_x_master, origin_y_master, origin_z_master, vecyx, vecyy, vecyz, color="g", **kwargs)
    ax.quiver(origin_x_master, origin_y_master, origin_z_master, veczx, veczy, veczz, color="b", **kwargs)
    # ax.axis('equal')

    return ax


def plot_points(points: Points | list[Point], master=None, fig_name: str | None = None, **kwargs):
    """Plots the given collection of points.

    Use ax.set_xlim3d(min, max) to properly set the ranges of the display.

    Args:
        points (Points | list[Point]): Collection of points to plot.
        master (ReferenceFrame | None): Optional master reference frame.
        fig_name (str): Name of the plot figure.
        kwargs : Keyword arguments passed to matplotlib.axes._subplots.Axes3DSubplot.scatter.

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the given collection of points
    """

    if master is None:
        temp_master = ReferenceFrame.create_master()
    else:
        temp_master = master.__copy__()

    if isinstance(points, list):
        all_points = Points(points, reference_frame=temp_master)
    elif isinstance(points, Points) or isinstance(points, egse.coordinates.point.Points):
        all_points = points
    else:
        raise ValueError("If the input is a list, all items in it must be Point objects")

    del temp_master

    coordinates = all_points.coordinates
    xs = coordinates[0, :]
    ys = coordinates[1, :]
    zs = coordinates[2, :]

    kwargs.setdefault("s", 50)
    kwargs.setdefault("marker", "o")
    kwargs.setdefault("color", "k")

    fig = plt.figure(fig_name)
    ax = fig.add_subplot(projection="3d")
    ax.scatter(xs, ys, zs, **kwargs)

    return ax


def plot_vectors(
    points: Points | list[Point], master=None, fig_name: str | None = None, from_origin: bool = True, **kwargs
):
    """Plots the given collection of vectors.

    Use ax.set_xlim3d(min, max) to properly set the ranges of the display.

    Args:
        points (Points | list[Point]): Collection of vectors to plot.
        master (ReferenceFrame | None): Optional master reference frame.
        fig_name (str): Name of the plot figure.
        kwargs : Keyword arguments passed to matplotlib.axes._subplots.Axes3DSubplot.scatter.

    Returns:
        matplotlib.axes._subplots.Axes3DSubplot displaying the given collection of points
    """

    if master is None:
        temp_master = ReferenceFrame.create_master()
    else:
        temp_master = master.__copy__()

    if isinstance(points, list):
        all_points = Points(points, reference_frame=temp_master)
    elif isinstance(points, Points) or isinstance(points, egse.coordinates.point.Points):
        all_points = points
    else:
        raise ValueError("If the input is a list, all items in it must be Point objects")

    del temp_master

    kwargs.setdefault("color", "k")

    # Prepare vector coordinates

    coordinates = all_points.coordinates
    xs = coordinates[0, :]
    ys = coordinates[1, :]
    zs = coordinates[2, :]

    # Origin of the x, y, and z vectors
    # -> x = The x coordinates of the origin of all vectors
    # -> [x,y,z] = Origin of points.reference_frame
    x, y, z = points.reference_frame.get_origin().coordinates[:3]
    x = np.ones_like(xs) * x
    y = np.ones_like(xs) * y
    z = np.ones_like(xs) * z

    # Plot

    fig = plt.figure(fig_name)
    ax = fig.gca(projection="3d")

    if from_origin:
        ax.quiver(x, y, z, xs - x, ys - y, zs - z, **kwargs)

    elif not from_origin:
        ax.quiver(xs, ys, zs, x - xs, y - ys, z - zs, **kwargs)

    else:
        print("Parameter 'from_origin' must be True or False")
        print("Setting it to True by default")
        ax.quiver(x, y, z, xs - x, ys - y, zs - z, **kwargs)

    return ax



from egse.scpi import count_number_of_channels
from egse.scpi import create_channel_list
from egse.scpi import get_channel_names


def test_get_channel_names():
    assert get_channel_names("1") == []
    assert get_channel_names("(@101)") == ["101"]
    assert get_channel_names("(@101,102,103)") == ["101", "102", "103"]
    assert get_channel_names("(@101:105)") == ["101", "102", "103", "104", "105"]
    assert get_channel_names("(@101, 102, 103, 105)") == ["101", "102", "103", "105"]
    assert get_channel_names("(@201:204,205,206,207:209,211)") == [
        "201",
        "202",
        "203",
        "204",
        "205",
        "206",
        "207",
        "208",
        "209",
        "211",
    ]


def test_get_channel_names_invalid():
    assert get_channel_names("") == []
    assert get_channel_names("101") == []
    assert get_channel_names("(@)") == []
    assert get_channel_names("(@101,,103)") == []
    assert get_channel_names("(@101:abc)") == []
    assert get_channel_names("(@101:102:103)") == []
    assert get_channel_names("(@101-105)") == []
    assert get_channel_names("(@10a)") == []
    assert get_channel_names("(@101, 10b)") == []
    assert get_channel_names("(@101:10c)") == []


def test_count_number_of_channels():
    assert count_number_of_channels("(@101)") == 1
    assert count_number_of_channels("(@101,102,103)") == 3
    assert count_number_of_channels("(@101:105)") == 5
    assert count_number_of_channels("(@101, 102, 103, 105)") == 4
    assert count_number_of_channels("(@201:204,205,206,207:209,211)") == 10
    assert count_number_of_channels("") == 0
    assert count_number_of_channels("101") == 0
    assert count_number_of_channels("(@)") == 0
    assert count_number_of_channels("(@101,,103)") == 0
    assert count_number_of_channels("(@101:abc)") == 0
    assert count_number_of_channels("(@101:102:103)") == 0
    assert count_number_of_channels("(@101-105)") == 0
    assert count_number_of_channels("(@10a)") == 0
    assert count_number_of_channels("(@101, 10b)") == 0
    assert count_number_of_channels("(@101:10c)") == 0


def test_create_channel_list():
    assert create_channel_list() == ""
    assert create_channel_list("101") == "(@101)"
    assert create_channel_list(["101"]) == "(@101)"
    assert create_channel_list(["101", "104"]) == "(@101:104)"
    assert create_channel_list("101", "102", "103") == "(@101,102,103)"
    assert create_channel_list("101", "102", "103", "104", "105") == "(@101,102,103,104,105)"
    assert create_channel_list("101", "102", "103", "105") == "(@101,102,103,105)"
    assert create_channel_list(["201", "204"], "205", "206", ["207", "209"], "211") == "(@201:204,205,206,207:209,211)"



import logging
import multiprocessing
from typing import Annotated, Optional

import rich
import sys
import typer
import zmq

from egse.control import is_control_server_active, ControlServer
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.settings import Settings
from egse.storage import store_housekeeping_information
from egse.tempcontrol.lakeshore.lakeshore336 import LakeShore336Proxy
from egse.tempcontrol.lakeshore.lakeshore336_protocol import LakeShore336Protocol
from egse.zmq_ser import connect_address, get_port_number

multiprocessing.current_process().name = "lakeshore336_cs"
CTRL_SETTINGS = Settings.load("LakeShore336 Control Server")

logger = logging.getLogger(__name__)


def is_lakeshore336_cs_active(device_id: str, timeout: float = 0.5) -> bool:
    """Checks if the LakeShore336 Control Server with the given device ID is active.

    Args
        device_id (str): Device identifier

    Returns: True if the LakeShore336 Control Server with the given device ID is active; False otherwise.
    """

    protocol = CTRL_SETTINGS.PROTOCOL
    hostname = CTRL_SETTINGS.HOSTNAME
    port = CTRL_SETTINGS[device_id]["COMMANDING_PORT"]

    endpoint = connect_address(protocol, hostname, port)

    return is_control_server_active(endpoint, timeout)


class LakeShore336ControlServer(ControlServer):
    def __init__(self, device_id: str, simulator: bool = False):
        """Initialisation of a LakeShore336 Control Server.

        Args:
            device_id (str): Device identifier
            simulator (bool): Indicates whether the LakeShore336 Control Server should be started in simulator mode
        """

        super().__init__(device_id=device_id)

        self.device_protocol = LakeShore336Protocol(self, device_id, simulator=simulator)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type=f"{device_id}")

    def get_communication_protocol(self) -> str:
        """Returns the communication protocol used by the Control Server.

        Returns Communication protocol used by the Control Server, as specified in the settings.
        """

        return "tcp"

    def get_commanding_port(self):
        """Returns the commanding port used by the Control Server.

        Returns: Commanding port used by the Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_cmd_sock) or 0

    def get_service_port(self):
        """Returns the service port used by the Control Server.

        Returns: Service port used by the Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_service_sock) or 0

    def get_monitoring_port(self) -> int:
        """Returns the monitoring port used by the Control Server.

        Returns: Monitoring port used by the Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_mon_sock) or 0

    def get_storage_mnemonic(self) -> str:
        """Returns the storage mnemonic used by the Control Server.

        This is a string that will appear in the filename with the housekeeping information of the device, as a way of
        identifying the device.  If this is not included in the settings, the device identifier is used.


        Returns: Storage mnemonic used by the Control Server, as specified in the settings.  If not specified in the
                 settings, the device identifier will be used.
        """

        try:
            return CTRL_SETTINGS[self.device_id]["STORAGE_MNEMONIC"]
        except AttributeError:
            return self.device_id

    def is_storage_manager_active(self):
        """Checks whether the Storage Manager is active.

        Returns: True if the Storage Manager is active; False otherwise.
        """

        from egse.storage import is_storage_manager_active

        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Stores the given housekeeping information in a dedicated file.

        Args:
            data (dict): Dictionary containing parameter name and value of all device housekeeping.
        """

        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        """Registers the Control Server to the Storage Manager."""

        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            },
        )

    def after_serve(self) -> None:
        self.deregister_service()

    def unregister_from_storage_manager(self):
        """Unregisters the Control Server from the Storage Manager."""

        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())


app = typer.Typer()


@app.command()
def start(
    device_id: Annotated[str, typer.Argument(help="Device identifier, identifies the hardware controller")],
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="Start the LakeShore336 Control Server in simulator mode")
    ] = False,
):
    """Start the LakeShore336 Control Server for the given device ID.

    Args:
        device_id (str): Device identifier
        simulator (bool): Indicates whether the LakeShore336 Control Server should be started in simulator mode
    """

    try:
        controller = LakeShore336ControlServer(device_id, simulator)
        controller.serve()

    except KeyboardInterrupt:
        print("Shutdown requested...exiting")

    except SystemExit as exit_code:
        print("System Exit with code {}.".format(exit_code))
        sys.exit(exit_code.code)

    except Exception:
        logger.exception(f"Cannot start the LakeShore336 Control Server {device_id}")

    return 0


@app.command()
def stop(device_id: str):
    """Stops the LakeShore336 Control Server for the given device ID.

    Args:
        device_id (str): Device identifier
    """

    with RegistryClient() as reg:
        service = reg.discover_service(device_id)
        rich.print("service = ", service)

        if service:
            proxy = ServiceProxy(protocol="tcp", hostname=service["host"], port=service["metadata"]["service_port"])
            proxy.quit_server()


@app.command()
def status(device_id: str):
    """Returns the status of the LakeShore336 Control Server for the given device ID.

    Args:
        device_id (str): Device identifier
    """

    with RegistryClient() as reg:
        service = reg.discover_service(device_id)

        if service:
            protocol = service.get("protocol", "tcp")
            hostname = service["host"]
            port = service["port"]
            service_port = service["metadata"]["service_port"]
            monitoring_port = service["metadata"]["monitoring_port"]
            endpoint = connect_address(protocol, hostname, port)

            if is_control_server_active(endpoint):
                rich.print(f"{device_id} CS: [green]active")
                with LakeShore336Proxy(device_id) as proxy:
                    sim = proxy.is_simulator()
                    connected = proxy.is_connected()
                    ip = proxy.get_ip_address()
                    rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
                    rich.print(f"hostname: {ip}")
                    rich.print(f"commanding port: {port}")
                    rich.print(f"service port: {service_port}")
                    rich.print(f"monitoring port: {monitoring_port}")
            else:
                rich.print(f"LakeShore336 {device_id} CS: [red]inactive")
        else:
            rich.print(
                f"[red]The LakeShore336 CS '{device_id}' isn't registered as a service. I cannot contact the control "
                f"server without the required info from the service registry.[/]"
            )
            rich.print(f"LakeShore336 {device_id}: [red]inactive")



"""
The Process Manager Control Server is the service that handles all Common-EGSE
processes.  The list of processes to handle, is taken from the Configuration
Manager (from the setup).

The Process Manager Control Server is implemented as a standard control server.
"""

import multiprocessing

import rich
import sys
import typer
import zmq
from rich.console import Console

from egse.control import ControlServer
from egse.log import logging
from egse.logger import remote_logging
from egse.process import SubProcess
from egse.procman import (
    PROCESS_NAME,
    SERVICE_TYPE,
    PROTOCOL,
    COMMANDING_PORT,
    SERVICE_PORT,
    MONITORING_PORT,
    HOSTNAME,
    STORAGE_MNEMONIC,
)
from egse.procman.procman_protocol import ProcessManagerProtocol
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.storage import store_housekeeping_information
from egse.zmq_ser import get_port_number

# Use explicit name here otherwise the logger will probably be called __main__

logger = logging.getLogger("egse.procman")


class ProcessManagerControlServer(ControlServer):
    def __init__(self):
        super().__init__()

        multiprocessing.current_process().name = PROCESS_NAME

        self.logger = logger
        self.service_name = PROCESS_NAME
        self.service_type = SERVICE_TYPE

        self.device_protocol = ProcessManagerProtocol(self)

        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type=SERVICE_TYPE)

        self.set_hk_delay(10.0)

        self.logger.info(f"PM housekeeping saved every {self.hk_delay / 1000:.1f} seconds.")

    def get_communication_protocol(self):
        return PROTOCOL

    def get_commanding_port(self):
        return get_port_number(self.dev_ctrl_cmd_sock) or COMMANDING_PORT

    def get_service_port(self):
        return get_port_number(self.dev_ctrl_service_sock) or SERVICE_PORT

    def get_monitoring_port(self):
        return get_port_number(self.dev_ctrl_mon_sock) or MONITORING_PORT

    def get_storage_mnemonic(self):
        return STORAGE_MNEMONIC

    def is_storage_manager_active(self):
        from egse.storage import is_storage_manager_active

        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            },
        )

    def unregister_from_storage_manager(self):
        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def after_serve(self):
        self.deregister_service()


app = typer.Typer(name=PROCESS_NAME)

console = Console(width=120)


@app.command()
def start():
    """Starts the Process Manager (pm_cs).

    The pm_cs is normally started automatically on egse-server boot.
    """

    multiprocessing.current_process().name = PROCESS_NAME

    with remote_logging():
        try:
            control_server = ProcessManagerControlServer()
            control_server.serve()
        except KeyboardInterrupt:
            print("Shutdown requested...exiting")
        except SystemExit as exit_code:
            print(f"System Exit with code {exit_code}.")
            sys.exit(exit_code.code)
        except Exception:
            import traceback

            traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def start_bg():
    """Starts the Process Manager Control Server in the background."""

    proc = SubProcess("pm_cs", ["pm_cs", "start"])
    proc.execute()


@app.command()
def stop():
    """Sends a 'quit_server' command to the Process Manager."""

    if COMMANDING_PORT == 0:
        with RegistryClient() as reg:
            service = reg.discover_service(SERVICE_TYPE)
            rich.print("service = ", service)
            if service:
                hostname = service["host"]
                port = service["metadata"]["service_port"]
            else:
                rich.print(
                    "[red]ERROR: Couldn't determine how to connect to the process manager. No service defined.[/]"
                )
                return
    else:
        hostname = HOSTNAME
        port = SERVICE_PORT

    rich.print("[green]Sending 'quit' command to process manager..[/]")
    try:
        with ServiceProxy(hostname=hostname, port=port) as proxy:
            proxy.quit_server()
    except ConnectionError as exc:
        console.print(f"[red]ERROR: Couldn't connect to process manager: {exc}[/]")


@app.command()
def status():
    """Prints the status of the control server."""

    import rich
    from egse.procman import get_status

    rich.print(get_status(), end="")


if __name__ == "__main__":
    sys.exit(app())



from enum import IntEnum

from egse.device import DeviceInterface
from egse.mixin import dynamic_command, CommandType, add_lf
from egse.proxy import Proxy
from egse.randomwalk import RandomWalk
from egse.registry.client import RegistryClient
from egse.settings import Settings
from egse.tempcontrol.lakeshore.lakeshore336_devif import LakeShore336EthernetInterface
from egse.zmq_ser import connect_address

CTRL_SETTINGS = Settings.load("LakeShore336 Control Server")


def split_response(response) -> list[float]:
    return response.split(",")


class SelfTestResult(IntEnum):
    """Possible results of the selftest."""

    NO_ERRORS_FOUND = 0
    ERRORS_FOUND = 1


class AutotuneMode(IntEnum):
    """Possible autotune modes."""

    PROPORTIONAL = P = 0  # P only
    PROPORTIONAL_INTEGRAL = PI = 1  # P and I
    PROPORTIONAL_INTEGRAL_DERIVATIVE = PID = 2  # P, I, and D


class HeaterQuantity(IntEnum):
    """Possible heater quantities."""

    CURRENT = 1
    POWER = 2


class HeaterRange(IntEnum):
    """Possible heater ranges."""

    OFF = 0  # Output 1 - 4
    LOW = 1  # Output 1 & 2
    MEDIUM = 2  # Output 1 & 2
    HIGH = 3  # Output 1 & 2

    ON = 1  # Output 3 & 4


class Mode(IntEnum):
    """Possible operating modes."""

    LOCAL = 0
    REMOTE = 1
    REMOTE_WITH_LOCAL_LOCKOUT = 2


class DisplayMode(IntEnum):
    """Possible display modes."""

    INPUT_A = 0
    INPUT_B = 1
    INPUT_C = 2
    INPUT_D = 3
    CUSTOM = 4
    FOUR_LOOP = 5
    ALL_INPUTS = 6


class SensorType(IntEnum):
    """Possible sensor types."""

    DISABLED = 0
    DIODE = 1  # 3062 option only
    PTC_RTD = 2
    NTC_RTD = 3
    THERMOCOUPLE = 4  # 3060 option only
    CAPACITANCE = 5  # 3060 option only


class InputSimulator:
    def __init__(self, input: str) -> None:
        self.input = input
        self.name = None
        self.type = None
        self.temperature_limit = None
        self.min_temperature = None
        self.max_temperature = None

        self.randomwalk = RandomWalk(start=15, boundary=(-100, 25), scale=0.01, count=0)

    def get_temperature(self):
        return next(self.randomwalk)


class OutputSimulator:
    def __init__(self, output: int) -> None:
        self.output = output
        self.resistance_setting = None
        self.max_current_setting = None
        self.max_user_current = None
        self.heater_quantity = None

        self.proportional = self.p = None
        self.integral = self.i = None
        self.derivative = self.d = None

        self.temperature_setpoint = None
        self.heater_range = None

    def set_pid_parameters(self, p, i, d):
        self.p = p
        self.i = i
        self.d = d

    def get_pid_parameters(self):
        return self.p, self.i, self.d


class LakeShore336Interface(DeviceInterface):
    """Interface for the LakeShore336 Controller, Simulator, and Proxy."""

    def __init__(self, device_id: str):
        """Initialisation of a LakeShore336 interface.

        Args:
            device_id (str): Device identifier
        """

        super().__init__()

        self.device_id = device_id

    @dynamic_command(cmd_type=CommandType.READ, cmd_string="*IDN?", process_cmd_string=add_lf)
    def info(self) -> (str, str, str, float):
        """Identification query.

        Returns basic information about the device.

        Returns: Tuple with the following information:
            - Manufacturer identifier (e.g. "LSCI");
            - Instrument model number (e.g. "MODEL336");
            - Instrument serial number / Option card serial number (e.g. "1234567/1234567");
            - Instrument firmware version (e.g. 1.0)
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="*CLS", process_cmd_string=add_lf)
    def clear_interface(self):
        """Clear interface command.

        Clears the bits in the status byte register, standard event status register, and operation event register, and
        terminates all pending operations.  Clears the interface but not the controller.  The related controller
        command is `reset_instrument`."""

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="*RST", process_cmd_string=add_lf)
    def reset_instrument(self):
        """Reset instrument command.

        Sets the controller parameters to power-up settings.
        """

        raise NotImplementedError

    # # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="*ESE", process_cmd_string=add_lf)
    #
    # # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="*ESE?", process_cmd_string=add_lf)
    #
    # # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="*ESR?", process_cmd_string=add_lf)

    @dynamic_command(cmd_type=CommandType.READ, cmd_string="TST?", process_cmd_string=add_lf)
    def get_selftest_result(self) -> int:
        """Selftest query.

        Reports status based on test done at power-up.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="ATUNE ${output_channel}, ${mode}")
    def autotune(self, output_channel: int, mode: AutotuneMode):
        """Autotune command.

        Configures autotune parameters.

        Args:
            output_channel (int): Output associated with the loop to be autotuned (1 - 4)
            mode (int): Autotune mode to start using:
                - 0 (AutotuneMode.PROPORTIONAL / AutotuneMode.P): Autotune in proportional mode;
                - 1 (AutotuneMode.PROPORTIONAL_INTEGRAL / AutotuneMode.PI): Autotune in proportional and integral mode;
                - 2 (AutotuneMode.PROPORTIONAL_INTEGRAL_DERIVATIVE / AutotuneMode.PID): Autotune in proportional,
                  integral, and derivative mode.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.READ, cmd_string="TUNEST?", process_cmd_string=add_lf)
    def get_tuning_status(self):
        """Control tuning status query.

        Returns:
            Tuning status (int):
                - 0: No active tuning
                - 1: Active tuning
            Heater output (int) of the control loop being tuned (when tuning)
                - 1: Heater output 1
                - 2: Heater output 2
            Error status (int):
                - 0: No tuning error
                - 1: Tuning error
            Current stage in the autotuning process that failed, if any (when tuning)

        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.READ, cmd_string="CRDG? {input_channel}", process_cmd_string=add_lf)
    def get_temperature(self, input_channel: str):
        """Celsius reading query.

        Args:
            input_channel (str): Input channel for which to return the current temperature

        Returns the current temperature reading for the specified input channel.
        """

        raise NotImplementedError

    @dynamic_command(cmd_type=CommandType.READ, cmd_string="HTR? ${output_channel}", process_cmd_string=add_lf)
    def get_heater_output(self, output_channel: int):
        """Heater output query.

        Returns the heater output.

        Args:
            output_channel (int): Heater output to query (1 or 2)

        Returns: Heater output in percentage.
        """

        raise NotImplementedError

    #
    # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="HTRSET?", process_cmd_string=add_lf)
    # # def get_heater_setup(self, output: int) -> (int, int, float, int):
    # #     """ Returns the heater setup.
    # #
    # #     Args:
    # #         output (int): Specifies heater output to query (1 or 2)
    # #
    # #     Returns: Heater setup:
    # #         - Heater resistance setting (int):
    # #             - 1: 25 Ohm;
    # #             - 2: 50 Ohm.
    # #         - Maximum heater output setting (int):
    # #             - 0: user-specified;
    # #             - 1: 0.707A;
    # #             - 2: 1A;
    # #             - 3: 1.141A;
    # #             - 4: 2A.
    # #         - Maximum user current (float)
    # #         - Heater quantity, specifying whether the heater output should be displayed in current
    # #           (HeaterQuantity.CURRENT / 1) or in power (HeaterQuantity.POWER / 2)
    # #
    # #     """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="HTRST?", process_cmd_string=add_lf)
    # # def get_heater_status(self, output: int) -> int:
    # #     """ Returns the heater status for the given output.
    # #
    # #     The error condition is cleared upon querying the heater status, which will also clear the error message on the
    # #     front panel.
    # #
    # #     Args:
    # #         output (int): Specifies heater output to query (1 or 2)
    # #
    # #     Returns: Heater status for the given output:
    # #         - 0: No error;
    # #         - 1: Heater open load;
    # #         - 2: Heater short.
    # #     """
    # #
    # #     raise NotImplementedError
    # # # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="*SRE", process_cmd_string=add_lf())
    # #
    # # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="*SRE?", process_cmd_string=add_lf())
    # #
    # # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="MODE ${mode}", process_cmd_string=add_lf)
    # # def set_mode(self, mode: Mode):
    # #     """ Places the controller in the given mode.
    # #
    # #     Args:
    # #         mode (Mode): Interface mode:
    # #             - 0 (Mode.LOCAL): Local mode;
    # #             - 1 (Mode.REMOTE): Remote mode;
    # #             - 2 (Mode.REMOTE_WITH_LOCAL_LOCKOUT): Remote mode with local lockout.
    # #     """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="MODE?", process_cmd_string=add_lf)
    # # def get_mode(self) -> int:
    # #     """ Returns the mode of the controller.
    # #
    # #     Returns: Mode of the controller:
    # #         - 0 (Mode.LOCAL): Local mode;
    # #         - 1 (Mode.REMOTE): Remote mode;
    # #         - 2 (Mode.REMOTE_WITH_LOCAL_LOCKOUT): Remote mode with local lockout.
    # #     """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="MOUT ${output}, ${value}", process_cmd_string=add_lf)
    # # def set_manual_output(self, output: int, value: float):
    # #     """ Sets the output value for the given output.
    # #
    # #     Args:
    # #         output (int): Specifies heater output
    # #         value (float): New value for the given heater output.
    # #     """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="MOUT?", process_cmd_string=add_lf)
    # # def get_manual_output(self, output: int):
    # #     """ Returns the output value for the given output.
    # #
    # #     Args:
    # #         output (int): Specifies heater output
    # #
    # #     Returns: Output value for the given output.
    # #    """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="DISPLAY ${mode}, ${num_fields}, ${output_source}", process_cmd_string=add_lf)
    # # def set_display_setup(self, mode: DisplayMode, num_fields: int, output_source: int):
    # #     """ Specifies the display mode.
    # #
    # #     Args:
    # #         mode (DisplayMode): Display mode:
    # #             - 0 (DisplayMode.INPUT_A);
    # #             - 1 (DisplayMode.INPUT_B);
    # #             - 2 (DisplayMode.INPUT_C);
    # #             - 3 (DisplayMode.INPUT_D);
    # #             - 4 (DisplayMode.CUSTOM);
    # #             - 5 (DisplayMode.FOUR_LOOP);
    # #             - 6 (DisplayMode.ALL_INPUTS)
    # #         num_fields (int): When `mode` is set to DisplayMode.CUSTOM, specifies the number of fields (display
    # #                           locations) to display:
    # #                                 - 0: 2 large;
    # #                                 - 1: 4 large;
    # #                                 - 2: 8 small.
    # #                           When `mode` is set to DisplayMode.All_INPUTS, specifies the size of hte readings:
    # #                                 - 0: Small with input names;
    # #                                 - 1: Large without input names.
    # #                           Ignored in all display modes (`mode`) except DisplayMode.CUSTOM.
    # #
    # #         output_source (int): Specifies which output and associated loop information to display in the bottom half
    # #                              of the custom display screen:
    # #                                 - 1: Output 1;
    # #                                 - 2: Output 2;
    # #                                 - 3: Output 3;
    # #                                 - 4: Output 4.
    # #                              Ignored in all display modes (`mode`) except DisplayMode.CUSTOM.
    # #     """
    # #
    # #     raise NotImplementedError
    # #
    # # @dynamic_command(cmd_type=CommandType.READ, cmd_string="DISPLAY?", process_cmd_string=add_lf)
    # # def get_display_setup(self):
    # #     """ Returns the display setup.
    # #
    # #
    # #     """
    # #
    # #     raise NotImplementedError
    #
    # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="TLIMIT ${input}, ${limit}", process_cmd_string=add_lf)
    # def set_temperature_limit(self, input: str, limit: float):
    #     """ Sets temperature limit for the given input.
    #
    #     Args:
    #         input (str): Specifies which input to configure
    #         limit (float): Temperature limit [K] for which to shut down all control units when exceeded.  A temperature
    #                        limit of zero turns the temperature limit feature off for the given sensor input.
    #
    #     """
    #
    #     raise NotImplementedError
    #
    # @dynamic_command(cmd_type=CommandType.READ, cmd_string="TLIMIT? ${input}", process_cmd_string=add_lf)
    # def get_temperature_limit(self, input: str):
    #     """ Returns the temperature limit for the given input.
    #
    #     Args:
    #         input (str): Specifies for which input to return the temperature
    #
    #     Returns: Temperature limit for the given input [K]
    #     """
    #
    #     raise NotImplementedError
    #
    #
    # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="INNAME ${input}, ${name}", process_cmd_string=add_lf)
    # def set_input_name(self, input: str, name: str):
    #
    #     raise NotImplementedError
    #
    # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="INNAME? ${input}", process_cmd_string=add_lf)
    # def get_input_name(self, input: str):
    #
    #     raise NotImplementedError
    #
    #
    # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="HTRSET ${output}, ${resistance_setting}, ${max_current_setting}, ${max_user_current}, ${heater_quantity}", process_cmd_string=add_lf)
    # def set_heater_setup(self, output: int, resistance_setting: int, max_current_setting: float, max_user_current: float, heater_quantity: HeaterQuantity):
    #     """ Sets the heater setpoint.
    #
    #     Args:
    #         output (int): Heater output to configure (1 or 2)
    #         resistance_setting (int): Heater resistance setting
    #             - 1: 25 Ohm;
    #             - 2: 50 Ohm.
    #         max_current_setting (int): Specifies the maximum heater output current:
    #             - 0: user-specified;
    #             - 1: 0.707A;
    #             - 2: 1A;
    #             - 3: 1.141A;
    #             - 4: 2A.
    #         max_user_current (float): Maximum heater output if `max_current_setting` is set to 0 (user-specified)
    #         heater_quantity (HeaterQuantity): Specifies whether the heater output should be displayed in current
    #                                           (HeaterQuantity.CURRENT / 1) or in power (HeaterQuantity.POWER / 2)
    #     """
    #
    #     raise NotImplementedError

    # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="PID ${output}, ${proportional}, ${integral}, ${derivative}", process_cmd_string=add_lf)
    # def set_pid_parameters(self, output: int, proportional: float, integral: float, derivative: float):
    #     """ Control loop PID values query.
    #
    #
    #     """
    #
    #     raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.READ,
        cmd_string="PID? ${output_channel}",
        process_cmd_string=add_lf,
        post_cmd=split_response,
    )
    def get_pid_parameters(self, output_channel: int):
        """Control loop PID values query.

        Args:
            output_channel (int): Output channel for which to return the control loop PID parameters

        Returns: Control loop PID values for the given output channel.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.READ,
        cmd_string="SETP? {output_channel}",
        process_cmd_string=add_lf,
        post_cmd=split_response,
    )
    def get_temperature_setpoint(self, output_channel: int):
        """Control setpoint query.

        Args:
            output_channel (int): Output channel for which to return the control setpoint (1 - 4)

        Returns: Temperature setpoint for the given output channel [°C]
        """

    def quit(self):
        """Cleans up and stops threads that were started by the process."""

        pass

    # @dynamic_command(cmd_type=CommandType.TRANSACTION, cmd_string="SETP ${output}, ${temperature}")
    # def set_temperature_setpoint(self, output: int, temperature: float):
    #
    #     raise NotImplementedError
    #
    # @dynamic_command(cmd_type=CommandType.WRITE, cmd_string="RANGE ${output}, ${range}", process_cmd_string=add_lf)
    # def set_heater_range(self, output: int, range: int):
    #
    #     raise NotImplementedError
    #
    # @dynamic_command(cmd_type=CommandType.READ, cmd_string="RANGE? ${output}", process_cmd_string=add_lf)
    # def get_heater_range(self, output: int):
    #
    #     raise NotImplementedError


class LakeShore336Controller(LakeShore336Interface):
    def __init__(self, device_id: str):
        super().__init__(device_id)

        self.lakeshore = self.transport = LakeShore336EthernetInterface(device_id)

    def is_simulator(self) -> bool:
        return False

    def is_connected(self) -> bool:
        return self.lakeshore.is_connected()

    def connect(self):
        self.lakeshore.connect()

    def disconnect(self):
        self.lakeshore.disconnect()

    def reconnect(self):
        self.lakeshore.reconnect()

    def quit(self):
        # self.synoptics.disconnect_cs()

        pass


class LakeShore336Simulator(LakeShore336Interface):
    def __init__(self, device_id: str):
        super().__init__(device_id)

        self._is_connected = True

        self.manufacturer = "LakeShore"
        self.model = "LSCI336"
        self.instrument_serial_number = 1234567
        self.option_card_serial_number = 1234567
        self.firmware_version = 1.0

        self.input = {}
        for i in range(ord("A"), ord("D") + 1):
            self.input[chr(i)] = InputSimulator(chr(i))

        self.output = [None, OutputSimulator(1), OutputSimulator(2)]

    def is_simulator(self) -> bool:
        return True

    def is_connected(self) -> bool:
        return self._is_connected

    def connect(self):
        self._is_connected = True

    def disconnect(self):
        self._is_connected = False

    def reconnect(self):
        self._is_connected = True

    def info(self) -> (str, str, str, float):
        return (
            self.manufacturer,
            self.model,
            f"{self.instrument_serial_number}/{self.option_card_serial_number}",
            self.firmware_version,
        )

    def clear_interface(self):
        pass

    def reset_instrument(self):
        pass

    def get_selftest_result(self):
        # No errors found

        return SelfTestResult.NO_ERRORS_FOUND.value

    def autotune(self, output_channel: int, mode: AutotuneMode):
        pass

    def get_tuning_status(self):
        # TODO
        pass

    # def get_selftest_result(self):
    #
    #     return SelfTestResult.NO_ERRORS_FOUND
    #
    #
    # def set_input_name(self, input: str, name: str):
    #
    #     self.input[input].name = name
    #
    # def get_input_name(self, input: str):
    #
    #     return self.input[input].name
    #
    # def set_temperature_limit(self, input: str, limit: float):
    #
    #     self.input[input].temperature_limit = limit
    #
    # def get_temperature_limit(self, input: str):
    #
    #     return self.input[input].temperature_limit
    #
    # def set_heater_setup(self, output: int, resistance_setting: int, max_current_setting: float, max_user_current: float, heater_quantity: HeaterQuantity):
    #
    #     self.output[output].resistance_setting = resistance_setting
    #     self.output[output].max_current_setting = max_current_setting
    #     self.output[output].max_user_current = max_user_current
    #     self.output[output].heater_quantity = heater_quantity
    #
    # def set_pid_parameters(self, output: int, proportional: float, integral: float, derivative: float):
    #
    #     self.output[output].set_pid_parameters(proportional, integral, derivative)

    def set_pid_parameters(self, output: int, p: float, i: float, d: float):
        self.output[output].set_pid_parameters(p, i, d)

    def get_pid_parameters(self, output: int):
        return self.output[output].get_pid_parameters()

    # def set_temperature_setpoint(self, output: int, temperature: float):
    #
    #     self.output[output].temperature_setpoint = temperature

    def get_temperature_setpoint(self, output_channel: int):
        return self.output[output_channel].temperature_setpoint

    #
    # def set_heater_range(self, output: int, range: int):
    #
    #     self.output[output].heater_range = range
    #
    # def get_heater_range(self, output: int):
    #
    #     return self.output[output].heater_range

    def set_heater_output(self, output_channel: int, heater_output: float):
        self.output[output_channel].output = heater_output

    def get_heater_output(self, output_channel: int):
        return self.output[output_channel].output

    def get_temperature(self, input_channel: str):
        return self.input[input_channel].get_temperature()


class LakeShore336Proxy(Proxy, LakeShore336Interface):
    """Proxy to connect to a LakeShore 336 Control Server"""

    def __init__(self, device_id: str):
        """Proxy to connect to the LakeShore 336 with the given device identifier

        Args:
            device_id (str): Device identifier
        """

        with RegistryClient() as reg:
            service = reg.discover_service(device_id)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]
            else:
                raise RuntimeError(f"No service registered as {CTRL_SETTINGS.SERVICE_TYPE}")

        super().__init__(connect_address(protocol, hostname, port))



__all__ = [
    "show_processes",
]


import re

from egse.process import ProcessInfo
from egse.process import get_processes


def show_processes():
    """Returns of list of ProcessInfo data classes for matching processes from this package."""

    def filter_procs(pi: ProcessInfo):
        pattern = r"(lakeshore336)_(ui|cs|sim)"

        return re.search(pattern, pi.command)

    return get_processes(filter_procs)



"""Command protocol for the Digilent MEASURpoint DT8874."""

import logging
from pathlib import Path

from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.device import DeviceConnectionState
from egse.digilent.digilent import DigilentInterface
from egse.digilent.measurpoint.dt8874.dt8874 import Dt8874Simulator, Dt8874Controller
from egse.digilent.measurpoint.dt8874.dt8874_cs import ORIGIN
from egse.hk import read_conversion_dict, convert_hk_names
from egse.protocol import DynamicCommandProtocol
from egse.settings import Settings
from egse.setup import SetupError
from egse.system import format_datetime
from egse.zmq_ser import bind_address

_HERE = Path(__file__).parent
DEVICE_SETTINGS = Settings.load(filename="dt8874.yaml", location=_HERE)
LOGGER = logging.getLogger("egse.digilent.measurpoint.dt8874")


class Dt8874Command(ClientServerCommand):
    """Command class for the Digilent MEASURpoint DT8874 Control Server."""

    pass


class Dt8874Protocol(DynamicCommandProtocol):
    """Command protocol for the Digilent MEASURpoint DT8874 Control Server."""

    def __init__(self, control_server: ControlServer, simulator: bool = False):
        """Initialisation of a Digilent MEASURpoint DT8874 protocol.

        Args:
            control_server (ControlServer): Digilent MEASURpoint DT8874 Control Server.
            simulator (bool): Whether to use a simulator as the backend.
        """

        super().__init__(control_server)

        try:
            self.hk_conversion_table = read_conversion_dict(
                self.get_control_server().get_storage_mnemonic(), use_site=False
            )
        except SetupError:
            self.hk_conversion_table = None

        self.simulator = simulator

        if self.simulator:
            self.dt8874: DigilentInterface = Dt8874Simulator()
        else:
            self.dt8874: DigilentInterface = Dt8874Controller()

        try:
            self.dt8874.connect()
        except ConnectionError:
            LOGGER.warning("Couldn't establish connection to the Digilent MEASURpoint DT8874, check the log messages.")

        # self.metrics = define_metrics("DT8874")

    def get_bind_address(self) -> str:
        """Returns the bind address for the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Bind address for the Digilent MEASURpoint DT8874 Control Server.
        """

        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_device(self) -> DigilentInterface:
        """Returns the Digilent MEASURpoint DT8874 interface.

        Returns:
            Digilent MEASURpoint DT8874 interface.
        """

        return self.dt8874

    def get_status(self) -> dict:
        """Returns the status information for the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Status information for the Digilent MEASURpoint DT8874 Control Server.
        """

        status = super().get_status()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return status

        # TODO Add device-specific status information

        return status

    def get_housekeeping(self) -> dict:
        """Returns the housekeeping information for the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Housekeeping information for the Digilent MEASURpoint DT8874 Control Server.
        """

        result = dict()
        result["timestamp"] = format_datetime()

        if "RTD" in self.dt8874.channels:
            for rtd_type in self.dt8874.channels.RTD:
                rtd_temperatures = self.dt8874.get_rtd_temperature(
                    rtd_type=rtd_type, channels=self.dt8874.channels.RTD[rtd_type]
                )

                for channel_id, rtd_temperature in zip(self.dt8874.channel_lists.RTD[rtd_type], rtd_temperatures):
                    original_name = f"{ORIGIN}_T_RTD_{rtd_type}_CH{channel_id}"
                    result[original_name] = rtd_temperature

        if "THERMOCOUPLE" in self.dt8874.channels:
            for tc_type in self.dt8874.channels.THERMOCOUPLE:
                tc_temperatures = self.dt8874.get_thermocouple_temperature(
                    tc_type=tc_type, channels=self.dt8874.channels.THERMOCOUPLE[tc_type]
                )

                for channel_id, tc_temperature in zip(self.dt8874.channel_lists.THERMOCOUPLE[tc_type], tc_temperatures):
                    original_name = f"{ORIGIN}_T_TC_{tc_type}_CH{channel_id}"
                    result[original_name] = tc_temperature

        if "RESISTANCE" in self.dt8874.channels:
            resistances = self.dt8874.get_resistance(self.dt8874.channels.RESISTANCE)

            for channel_id, resistance in zip(self.dt8874.channel_lists.RESISTANCE, resistances):
                original_name = f"{ORIGIN}_R_CH{channel_id}"
                result[original_name] = resistance

        if "VOLTAGE" in self.dt8874.channels:
            voltages = self.dt8874.get_voltage(self.dt8874.channels.VOLTAGE)

            for channel_id, voltage in zip(self.dt8874.channel_lists.VOLTAGE, voltages):
                original_name = f"{ORIGIN}_V_CH{channel_id}"
                result[original_name] = voltage

        if self.hk_conversion_table:
            return convert_hk_names(result, self.hk_conversion_table)
        return result

    def is_device_connected(self) -> bool:
        """Checks whether the Digilent MEASURpoint DT8874 is connected.

        Returns:
            True if the Digilent MEASURpoint DT8874 is connected; False otherwise.
        """

        return self.dt8874.is_connected()



"""
Configure local logging for the CGSE.

I use 'local logging' here because the CGSE also has a Logger server which stores
all CGSE log messages in a rotated file. That server is part of the `cgse-core` package.

Environment variables that affect logging:

  - LOG_FORMAT: full | FULL
  - LOG_LEVEL: an integer [10, 50] or a level name DEBUG, INFO, WARNING, CRITICAL, ERROR

"""

__all__ = [
    "LOG_FORMAT_FULL",
    "LOG_FORMAT_CLEAN",
    "LOG_FORMAT_STYLE",
    "LOG_DATE_FORMAT_FULL",
    "LOG_DATE_FORMAT_CLEAN",
    "logger",
    "logging",  # export to guarantee that this module is imported and initialized before users use logging
    "root_logger",
    "egse_logger",
    "get_log_level_from_env",
    "PackageFilter",
]

import logging
import os
import textwrap
from pathlib import Path

import rich

# The format for the log messages.
# The log record attributes are listed: https://docs.python.org/3.12/library/logging.html#logrecord-attributes

LOG_FORMAT_STYLE = "{"
LOG_FORMAT_FULL = (
    "{asctime:19s}.{msecs:03.0f} : {processName:20s} : {levelname:8s} : {name:^25s} : {lineno:6d} : {filename:20s} : {"
    "message}"
)
LOG_FORMAT_CLEAN = (
    "{asctime} [{levelname:>8s}] {message} ({processName}[{process}]:{package_name}:{filename}:{lineno:d})"
)

LOG_DATE_FORMAT_FULL = "%Y-%m-%d %H:%M:%S"
LOG_DATE_FORMAT_CLEAN = "%Y-%m-%d %H:%M:%S"


class PackageFilter(logging.Filter):
    """Adds 'package_name' to the log record.

    When this filter is added to a handler of a logger, the formatter of that
    logger can use the 'package_name' attribute.

    When the package name can not be determined, is will contain 'n/a'.

    NOTE: this filer assumes the root package is 'egse'.
    """

    def filter(self, record):
        if hasattr(record, "pathname"):
            parts = Path(record.pathname).parent.parts
            try:
                egse_index = parts.index("egse")
                package_name = ".".join(parts[egse_index:])
            except ValueError:
                package_name = "n/a"

            record.package_name = package_name
        else:
            record.package_name = "n/a"

        return True


class EGSEFilter(logging.Filter):
    def filter(self, record):
        return record.name.startswith("egse")


class NonEGSEFilter(logging.Filter):
    def filter(self, record):
        return not record.name.startswith("egse")


def get_log_level_from_env(env_var: str = "LOG_LEVEL", default: str = "INFO") -> int:
    """Read the log level from an environment variable."""
    log_level_str = os.getenv(env_var, default)

    # Try to convert to integer first (for numeric levels)
    try:
        log_level = int(log_level_str)

        if 10 <= log_level <= 50:
            return log_level
        else:
            logging.warning(f"Log level {log_level} outside standard range (10-50). Using {default.upper()}.")
            return logging._nameToLevel[default.upper()]

    except ValueError:
        log_level_str = log_level_str.upper()
        try:
            return getattr(logging, log_level_str)
        except AttributeError:
            logging.error(f"Invalid LOG_LEVEL '{log_level_str}'. Using {default.upper()}.")
            return logging._nameToLevel[default.upper()]


egse_logger = logging.getLogger("egse")
egse_logger.level = get_log_level_from_env()  # We might want to choose another env e.g. CGSE_LOG_LEVEL

root_logger = logging.getLogger()
root_logger.level = get_log_level_from_env()

egse_handler = logging.StreamHandler()
if os.getenv("LOG_FORMAT", "").lower() == "full":
    egse_formatter = logging.Formatter(fmt=LOG_FORMAT_FULL, datefmt=LOG_DATE_FORMAT_FULL, style=LOG_FORMAT_STYLE)
else:
    egse_formatter = logging.Formatter(fmt=LOG_FORMAT_CLEAN, datefmt=LOG_DATE_FORMAT_CLEAN, style=LOG_FORMAT_STYLE)

egse_handler.setFormatter(egse_formatter)
egse_handler.addFilter(EGSEFilter())
egse_handler.addFilter(PackageFilter())

root_logger.addHandler(egse_handler)

for handler in root_logger.handlers:
    if handler != egse_handler:  # Don't filter our new handler
        handler.addFilter(NonEGSEFilter())
        handler.addFilter(PackageFilter())

# Optional: integrate with Textual logging if available
#
# try:
#     from textual.logging import TextualHandler

#     root_logger.addHandler(TextualHandler())
# except ImportError:
#     pass


logger = egse_logger

if __name__ == "__main__":
    from egse.env import str_env

    root_logger = logging.getLogger()

    print(
        textwrap.dedent(
            """
            Environment variables:
              - LOG_LEVEL=debug|info|warning|critical
              - LOG_FORMAT=full|clean

            Example logging statements
              - logging level set to INFO
              - logging format set to full
              - fields are separated by a colon ':'
              - fields: date & time: process name : level : logger name : lineno : filename : message
            """
        )
    )

    if str_env("LOG_FORMAT", "clean").strip().lower() == "full":
        rich.print(
            f"[b]{'Date & Time':^23s} : {'Process Name':20s} : {'Level':8s} : {'Logger Name':^25s} : {' Line '} : "
            f"{'Filename':20s} : {'Message'}[/]"
        )
    else:
        rich.print(f"[b]{'Date & Time':^19s} [ Level  ] Message (filename:lineno)[/]")

    rich.print("-" * 150)
    for name, level in logging.getLevelNamesMapping().items():
        logger.log(level, f"{name} logging message")

    root_logger.info("This should come out of the root logger, not the egse logger.")



import asyncio
import csv
import datetime
import logging
import operator
import os
import pprint
import random
import shutil
import textwrap
import time
from pathlib import Path
from unittest.mock import patch

import pytest
import rich
from fixtures.helpers import create_empty_file
from fixtures.helpers import create_text_file
from pytest import approx

from egse.decorators import execution_count
from egse.system import AttributeDict
from egse.system import Periodic
from egse.system import SignalCatcher
from egse.system import Timer
from egse.system import camel_to_kebab
from egse.system import camel_to_snake
from egse.system import check_argument_type
from egse.system import clear_average_execution_times
from egse.system import do_every
from egse.system import duration
from egse.system import env_var
from egse.system import execution_time
from egse.system import filter_by_attr
from egse.system import format_datetime
from egse.system import get_average_execution_time
from egse.system import get_average_execution_times
from egse.system import get_caller_breadcrumbs
from egse.system import get_caller_info
from egse.system import get_current_location
from egse.system import get_full_classname
from egse.system import get_os_name
from egse.system import get_os_version
from egse.system import get_referenced_var_name
from egse.system import get_system_name
from egse.system import get_system_stats
from egse.system import humanize_seconds
from egse.system import is_in
from egse.system import is_module
from egse.system import is_namespace
from egse.system import is_not_in
from egse.system import ping
from egse.system import read_last_line
from egse.system import read_last_lines
from egse.system import recursive_dict_update
from egse.system import redirect_output_to_log
from egse.system import replace_environment_variable
from egse.system import round_up
from egse.system import save_average_execution_time
from egse.system import touch
from egse.system import wait_until
from egse.system import waiting_for

HERE = Path(__file__).parent


def test_get_current_location():
    def internal_function():
        return get_current_location()

    location = internal_function()

    assert "test_system.py" in location[0]
    assert type(location[1]) is int
    assert "internal_function" in location[2]


def test_round_up():
    assert round_up(3.14159, 0) == 4.0
    assert round_up(3.14159, 1) == 3.2
    assert round_up(3.14159, 2) == 3.15
    assert round_up(3.14159, 3) == 3.142

    # an integer is returned as float, but no rounding is done
    assert round_up(4, 3) == 4.0
    assert round_up(1, 2) == 1.0


def test_attr_dict():
    ad = AttributeDict()
    assert not ad
    assert len(ad) == 0

    ad = AttributeDict({})
    assert not ad
    assert len(ad) == 0

    ad = AttributeDict({"a": 1, "b": 2, "c": 3})
    assert ad
    assert len(ad) == 3
    assert ad.a == 1
    assert ad.b == 2
    assert ad.c == 3
    assert ad["a"] == 1
    assert ad["b"] == 2
    assert ad["c"] == 3

    ad = AttributeDict()
    ad.d = 4
    ad.e = 5
    ad.f = 6
    assert len(ad) == 3
    assert ad["d"] == 4
    assert ad["e"] == 5
    assert ad["f"] == 6
    assert ad.d == 4
    assert ad.e == 5
    assert ad.f == 6

    ad = AttributeDict({"a": 1, "b": 2, "c": 3})

    assert hasattr(ad, "a") and "a" in ad
    assert hasattr(ad, "b") and "b" in ad
    assert hasattr(ad, "c") and "c" in ad

    assert not hasattr(ad, "d")
    assert "d" not in ad

    del ad["b"]

    assert not hasattr(ad, "b")
    assert "b" not in ad

    del ad.c

    assert not hasattr(ad, "c")
    assert "c" not in ad

    ad = AttributeDict()

    # It is allowed to add objects to the AttributeDict, but we can not use them as attributes

    ad[1] = "a"
    ad[(7, 3)] = 73

    assert 73 in ad.values()

    with pytest.raises(KeyError):
        assert ad["a"] == 2

    with pytest.raises(AttributeError):
        assert ad.a == "2"


def test_attr_dict_rich():
    import rich

    ad = AttributeDict({"a": 1, "b": 2, "c": 3})
    rich.print(ad)
    assert str(ad) == "AttributeDict({'a':1, 'b':2, 'c':3})"

    ad = AttributeDict({"a": 1, "b": {"B": 2, "BB": 22}, "c": 3}, label="nested dict")
    rich.print(ad)
    assert str(ad) == "AttributeDict({'a':1, 'b':{'B': 2, 'BB': 22}, 'c':3}, label='nested dict')"

    ad = AttributeDict(
        {
            "a": 1,
            "b": {"B": 2, "BB": 22},
            "c": 3,
            "d": 4,
            "e": 5,
            "f": 6,
            "g": 7,
            "h": 8,
            "i": 9,
            "j": 10,
            "k": 11,
            "l": 12,
        },
        label="Long nested label",
    )
    rich.print(ad)
    assert str(ad) == (
        "AttributeDict({'a':1, 'b':{'B': 2, 'BB': 22}, 'c':3, 'd':4, 'e':5, 'f':6, 'g':7, 'h':8, 'i':9, "
        "'j':10, ...}, label='Long nested label')"
    )


def test_get_call_sequence():
    assert get_caller_breadcrumbs().startswith("call stack: test_get_call_sequence[")
    assert len(get_caller_breadcrumbs().split(" <- ")) == 5

    assert get_caller_breadcrumbs(prefix="XXX: ").startswith("XXX: test_get")
    assert "<-" not in get_caller_breadcrumbs(limit=1)

    assert "test_system.py" in get_caller_breadcrumbs(with_filename=True)


def test_caller():
    assert "test_caller" in get_caller_info()
    assert "test_caller" == get_caller_info().function
    assert __file__ in get_caller_info()
    assert __file__ in get_caller_info().filename
    assert type(get_caller_info().lineno) is int

    assert get_caller_info(level=0).filename.endswith("/egse/system.py")
    assert "get_caller_info" == get_caller_info(level=0).function

    # boundary test, eventually we end up at the <module> level

    assert get_caller_info(level=50).function == "<module>"
    assert get_caller_info(level=250).function == "<module>"

    def internal_function():
        return get_caller_info()

    assert "internal_function" in internal_function()
    assert __file__ in internal_function()

    class MyClass:
        def a_function(self):
            return get_caller_info(level=1)

    assert "a_function" in MyClass().a_function()
    assert __file__ in MyClass().a_function()


def test_var_names():
    a_var = 42
    assert get_referenced_var_name(a_var) == ["a_var"]

    b_var = a_var  # noqa
    assert "b_var" in get_referenced_var_name(a_var)
    assert "a_var" in get_referenced_var_name(a_var)

    # A function is also an object...

    assert get_referenced_var_name(test_var_names) == ["test_var_names"]

    # Test the use of this function inside a class, e.g. for reporting

    class MyClass:
        def __str__(self):
            return ", ".join(get_referenced_var_name(self))

    my_class = this_class = MyClass()  # noqa

    assert "my_class" in f"{my_class}"
    assert "this_class" in f"{my_class}"

    # There is no reference to an 'anonymous' instance of MyClass

    assert "" == f"{MyClass()}"


def test_ping():
    assert ping("10.0.0.1") is False
    assert ping("127.0.0.1") is True
    assert ping("localhost") is True


def test_timer():
    with Timer("Testing Timer context manager", precision=4) as timer:
        time.sleep(1.0)
        timer.log_elapsed()
        timer.log_elapsed()
        time.sleep(1.0)
        timer.log_elapsed()
        timer.log_elapsed()

    # Confirm that the elapsed time is not increased any more

    elapsed = timer()
    assert elapsed > 2.0
    time.sleep(1.0)
    assert elapsed == timer()

    timer.log_elapsed()
    timer.log_elapsed()


def test_system_stats():
    with Timer():
        stats = get_system_stats()

    print("\nget_system_stats() = ", end="")
    pprint.pprint(stats)

    assert stats["cpu_count"] > 1
    assert stats["cpu_load"][2] < 100
    assert stats["total_ram"] > 8.0
    assert stats["avail_ram"] > 0.0


def test_wait_until():
    class SleepUntilCount:
        def __init__(self, end):
            self._end = end
            self._count = 0

        def __call__(self, *args, **kwargs):
            self._count += 1
            return self._count >= self._end

        def check_count(self):
            return self.__call__()

    assert wait_until(SleepUntilCount(10), interval=0.5) is True
    assert wait_until(SleepUntilCount(10), timeout=10) is False
    assert wait_until(SleepUntilCount(10).check_count, timeout=10) is False

    assert wait_until(lambda x: x, 0, interval=0.5, timeout=1) is True
    assert wait_until(lambda x: x, 1) is False


def test_waiting_for():
    class SleepUntilCount:
        def __init__(self, end):
            self._end = end
            self._count = 0

        def __call__(self, *args, **kwargs):
            self._count += 1
            return self._count >= self._end

        def check_count(self):
            return self.__call__()

    with pytest.raises(TimeoutError):
        waiting_for(SleepUntilCount(10), interval=0.5)

    assert waiting_for(SleepUntilCount(10), timeout=10)

    assert waiting_for(SleepUntilCount(10).check_count, timeout=10)

    with pytest.raises(TimeoutError):
        waiting_for(lambda x: x, 0, interval=0.5, timeout=1)

    assert waiting_for(lambda x: x, 1)


def test_os_name_and_version():
    # We test primarily for 'unknown' because if this happens, the code needs to be updated.

    assert get_os_name() != "unknown"
    assert get_os_name() in ("centos", "macos", "ubuntu")

    assert get_system_name() in ("linux", "darwin", "windows")

    # Here we test for version numbers of the os, which is expected to be up-to-date

    assert get_os_version() != "unknown"
    if get_os_name() == "macos":
        assert float(get_os_version()) > 10.14
    if get_os_name() == "centos":
        assert float(get_os_version()) >= 7.0


def test_humanize_seconds():
    assert humanize_seconds(0) == "00s.000"
    assert humanize_seconds(0.123) == "00s.123"
    assert humanize_seconds(53.43) == "53s.430"
    assert humanize_seconds(92) == "01m32s.000"
    assert humanize_seconds(92, include_micro_seconds=False) == "01m32s"

    assert humanize_seconds(60 * 60 + 120 + 3.002) == "01h02m03s.002"

    assert humanize_seconds(24 * 60 * 60 + 32.42) == "1d 32s.420"

    assert humanize_seconds(3 * 24 * 60 * 60 + 0 * 60 * 60 + 24 * 60 + 0.020) == "3d 24m00s.020"

    assert humanize_seconds(365 * 24 * 60 * 60 + 14 * 60 * 60 + 0 * 60 + 16 + 0.7) == "365d 14h00m16s.700"

    assert humanize_seconds(2351 * 24 * 60 * 60) == "2351d 00s.000"
    assert humanize_seconds(2351 * 24 * 60 * 60, include_micro_seconds=False) == "2351d 00s"

    with Timer("Timing 10_000 humanize_seconds() calls"):
        for d in range(10_000):
            humanize_seconds(d * 24 * 60 * 60 + 14 * 60 * 60 + 0 * 60 + 16 + 0.7)


def test_replace_environment_variable():
    assert replace_environment_variable("PLAIN_STRING") == "PLAIN_STRING"

    os.environ["DATA_STORAGE_LOCATION"] = "/Users/rik/data/CSL"
    assert replace_environment_variable("ENV['DATA_STORAGE_LOCATION']") == "/Users/rik/data/CSL"

    os.environ["DATA_STORAGE_LOCATION"] = "/Users/rik/data"
    assert replace_environment_variable("ENV['DATA_STORAGE_LOCATION']/CSL") == "/Users/rik/data/CSL"


class X:
    def __init__(self, d):
        for k, v in d.items():
            setattr(self, k, v)


def test_filter_by_attr():
    x0 = X({"a": 1, "b": 1})
    x1 = X({"a": 1, "b": 2})
    x2 = X({"a": 1, "b": 3, "c": 1})

    a_list = [x0, x1, x2]

    result = filter_by_attr(a_list, a=1)
    assert result == [x0, x1, x2]

    result = filter_by_attr(a_list, b=2)
    assert result == [x1]

    result = filter_by_attr(a_list, a=1, b=2)
    assert result == [x1]

    result = filter_by_attr(a_list, a=1, b=3, c=1)
    assert result == [x2]

    result = filter_by_attr(a_list, b=(is_in, (1, 2)))
    assert result == [x0, x1]

    result = filter_by_attr(a_list, b=(is_not_in, (1, 2)))
    assert result == [x2]

    result = filter_by_attr(a_list, b=(operator.ne, 2))
    assert result == [x0, x2]

    result = filter_by_attr(a_list, b=(operator.gt, 2))
    assert result == [x2]

    result = filter_by_attr(a_list, c=1, d=4)
    assert result == []

    result = filter_by_attr(a_list, d=4)
    assert result == []


def test_filter_by_attr_with_attribute_dict():
    a_list = [
        AttributeDict({"a": 1, "b": 1}),
        AttributeDict({"a": 1, "b": 2}),
        AttributeDict({"a": 1, "b": 3, "c": 1}),
    ]

    result = filter_by_attr(a_list, a=1)
    assert len(result) == 3

    result = filter_by_attr(a_list, b=2)
    assert len(result) == 1

    result = filter_by_attr(a_list, a=1, b=2)
    assert len(result) == 1

    result = filter_by_attr(a_list, a=1, b=3, c=1)
    assert len(result) == 1

    result = filter_by_attr(a_list, b=(lambda a, b: a in b, (1, 2)))
    assert len(result) == 2

    result = filter_by_attr(a_list, c=1, d=4)
    assert len(result) == 0

    result = filter_by_attr(a_list, d=4)
    assert len(result) == 0


def test_filter_by_attr_with_class():
    class X:
        def __init__(self, x, y, z):
            self.x = x
            self.y = y
            self.z = z

    class Y:
        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

    x0 = X(1, 2, 3)
    x1 = X(1, 3, 3)
    x2 = X(1, 2, 2)
    x3 = X(2, 3, 4)
    x4 = X(2, Y(1, 2, 3), 4)

    a_list = [x0, x1, x2, x3, x4]

    result = filter_by_attr(a_list, x=1)
    assert len(result) == 3

    result = filter_by_attr(a_list, y=2)
    assert len(result) == 2

    result = filter_by_attr(a_list, y__b=2)
    assert len(result) == 1

    result = filter_by_attr(a_list, y__a=(hasattr, True))
    assert len(result) == 1

    result = filter_by_attr(a_list, y__a=(hasattr, False))
    assert len(result) == 4

    x2 = X(1, Y(2, 3, 3), 3)
    a_list = [x0, x1, x2, x3, x4]

    result = filter_by_attr(a_list, y__b=(is_in, (1, 2)))
    assert len(result) == 1 and result == [x4]


def test_filter_by_attr_has():
    """Test if iterable has an attribute."""

    x0 = X({"a": 1, "b": 1, "d": 1})
    x1 = X({"a": 1, "b": 2, "d": 2})
    x2 = X({"a": 1, "b": 3, "c": 1, "d": 3})

    a_list = [x0, x1, x2]

    result = filter_by_attr(a_list, b=(hasattr, True))
    assert result == [x0, x1, x2]

    result = filter_by_attr(a_list, b=(hasattr, False))
    assert result == []

    result = filter_by_attr(a_list, c=(hasattr, True))
    assert result == [x2]

    result = filter_by_attr(a_list, c=(hasattr, False))
    assert result == [x0, x1]

    result = filter_by_attr(a_list, b=(hasattr, True), d=2, a=(operator.le, 1))
    assert result == [x1]

    result = filter_by_attr(a_list, b=(hasattr, True), d=2, a=(operator.lt, 1))
    assert result == []

    result = filter_by_attr(a_list, c=(hasattr, False), d=2, a=(operator.le, 1))
    assert result == [x1]


def test_filter_by_attr_has_with_attribute_dict():
    """Test if iterable has an attribute."""

    a_list = [
        AttributeDict({"a": 1, "b": 1}),
        AttributeDict({"a": 1, "b": 2}),
        AttributeDict({"a": 1, "b": 3, "c": 1}),
    ]

    result = filter_by_attr(a_list, b=(hasattr, True))
    assert len(result) == 3

    result = filter_by_attr(a_list, c=(hasattr, False))
    assert len(result) == 2


def test_read_last_line():
    # non-existing file

    line = read_last_line("/xxx")
    assert line is None

    prefix = Path(__file__).parent
    filename = prefix / "data/tmp/empty_data_file.txt"

    with create_empty_file(filename, create_folder=True):
        line = read_last_line(filename)
    assert line == ""

    filename = prefix / "data/tmp/data_file.txt"

    with create_text_file(
        filename,
        textwrap.dedent(
            """\
                001 002 003
                002 003 001
                003 002 001
                """
        ),
    ):
        line = read_last_line(filename)
    assert line == "003 002 001"


def test_read_last_lines():
    with pytest.raises(AssertionError, match="a positive number or zero"):
        _ = read_last_lines("/xxx", -1)

    assert read_last_lines("/xxx", 0) == []
    assert read_last_lines("/xxx", 1) == []

    prefix = Path(__file__).parent
    filename = prefix / "data/tmp/empty_data_file.txt"

    with create_empty_file(filename, create_folder=True):
        line = read_last_lines(filename, 3)
    assert line == []

    filename = prefix / "data/tmp/data_file.txt"

    with create_text_file(
        filename,
        textwrap.dedent(
            """\
                001 002 003
                002 003 001
                003 002 001
                """
        ),
    ):
        lines = read_last_lines(filename, 0)
        assert lines == [""]

        lines = read_last_lines(filename, 1)
        assert lines == ["003 002 001"]

        lines = read_last_lines(filename, 2)
        assert lines == ["002 003 001", "003 002 001"]

        lines = read_last_lines(filename, 10)
        assert lines == ["001 002 003", "002 003 001", "003 002 001"]


def test_read_last_lines_from_big_data_file(tmp_path):
    temp_file = tmp_path / "data.csv"

    fixed_data = generate_big_data_file(temp_file)

    with Timer(name="log-file", precision=6):
        lines = read_last_lines(temp_file, 5)
        assert len(lines) == 5

    assert lines == fixed_data[-5:]


def test_format_datetime():
    dt = datetime.datetime(2020, 6, 13, 14, 45, 45, 696138)

    dts = format_datetime(dt=dt)
    assert dts == "2020-06-13T14:45:45.696"

    dts = format_datetime(dt=dt, fmt="%Y-%m-%d %H:%M:%S")
    assert dts == "2020-06-13 14:45:45"

    dts = format_datetime(dt=dt, fmt="%y%m%d_%H%M%S")
    assert dts == "200613_144545"

    # Tests added to fix issue #1441 about incorrect timestamps

    dt = datetime.datetime(2020, 6, 13, 14, 45, 59, 999500)

    dts = format_datetime(dt=dt)
    assert dts == "2020-06-13T14:45:59.999"

    dts = format_datetime(dt=dt, precision=6)
    assert dts == "2020-06-13T14:45:59.999500"

    dt = datetime.datetime(2020, 6, 13, 14, 45, 59, 999501, tzinfo=datetime.timezone.utc)

    dts = format_datetime(dt=dt)
    assert dts == "2020-06-13T14:45:59.999+0000"

    dts = format_datetime(dt=dt, precision=6)
    assert dts == "2020-06-13T14:45:59.999501+0000"

    dts_today = format_datetime("today")
    dts_yesterday = format_datetime("yesterday")
    dts_day_before_yesterday = format_datetime("day before yesterday")
    dts_tomorrow = format_datetime("tomorrow")

    dt_today = datetime.datetime.strptime(dts_today, "%Y%m%d")
    dt_yesterday = datetime.datetime.strptime(dts_yesterday, "%Y%m%d")
    dt_day_before_yesterday = datetime.datetime.strptime(dts_day_before_yesterday, "%Y%m%d")
    dt_tomorrow = datetime.datetime.strptime(dts_tomorrow, "%Y%m%d")

    assert dt_today - dt_yesterday == datetime.timedelta(days=1)
    assert dt_today - dt_day_before_yesterday == datetime.timedelta(days=2)
    assert dt_today - dt_tomorrow == datetime.timedelta(days=-1)

    with pytest.raises(ValueError):
        _ = format_datetime("another day")

    dts = format_datetime()

    assert dts.startswith(format_datetime("today", fmt="%Y-%m-%d"))


def test_full_classname():
    assert get_full_classname(print) == "builtins.print"
    assert get_full_classname(int) == "builtins.int"
    assert get_full_classname(get_full_classname) == "egse.system.get_full_classname"
    assert get_full_classname(int(2)) == "builtins.int"
    assert get_full_classname(3.14) == "builtins.float"
    assert get_full_classname(3 + 6j) == "builtins.complex"
    assert get_full_classname(AttributeDict) == "egse.system.AttributeDict"
    assert get_full_classname(AttributeDict()) == "egse.system.AttributeDict"
    assert get_full_classname(datetime.datetime(2020, 6, 13, 14, 45, 45, 696138)) == "datetime.datetime"


def test_argument_type():
    check_argument_type("a string", "name", (int, str), allow_none=True)
    check_argument_type(None, "name", (int, str), allow_none=True)
    with pytest.raises(TypeError):
        check_argument_type(None, "name", (int, str), allow_none=False)
        print("Shouldn't have reached this line, check the code!")

    with pytest.raises(TypeError):
        check_argument_type(3.14, "name", (int, str), allow_none=True)
        print("Shouldn't have reached this line, check the code!")


def test_signals():
    killer = SignalCatcher()

    pid = os.getpid()

    os.kill(pid, 15)  # send a TERM signal to this process

    while not killer.term_signal_received:
        time.sleep(1.0)

    assert killer.signal_number == 15
    assert killer.signal_name == "SIGTERM"

    killer.clear(term=True)

    os.kill(pid, 30)  # send a SIGUSR1 signal to this process

    while not killer.user_signal_received:
        time.sleep(1.0)

    assert killer.signal_number == 30
    assert killer.signal_name == "SIGUSR1"

    killer.clear()


def test_env_var_context_manager():
    os.environ["TEST_ENV_VAR_1"] = "before context manager"

    assert os.environ.get("TEST_ENV_VAR_1") == "before context manager"
    assert os.environ.get("TEST_ENV_VAR_2") is None

    with env_var(TEST_ENV_VAR_1="within context manager", TEST_ENV_VAR_2="new variable"):
        assert os.environ.get("TEST_ENV_VAR_1") == "within context manager"
        assert os.environ.get("TEST_ENV_VAR_2") == "new variable"

    assert os.environ.get("TEST_ENV_VAR_1") == "before context manager"
    assert os.environ.get("TEST_ENV_VAR_2") is None


# The following is a slow test, it takes about 8s.
# You can deselect this test by running: `pytest -m 'not slow' src/tests/test_system.py`


@pytest.mark.slow
def test_function_timing():
    print()

    def one_second():
        time.sleep(1.0)

    def incr(count):
        time.sleep(count)
        return count + 1

    @execution_time
    def wait_seconds(wait_time: float):
        time.sleep(wait_time)

    x = 0
    for _ in range(2):
        x = save_average_execution_time(incr, x)

    assert get_average_execution_time(incr) == approx(0.5, abs=0.005, rel=0.005)

    for _ in range(10):
        save_average_execution_time(time.sleep, 0.2)

    assert get_average_execution_time(time.sleep) == approx(0.2, abs=0.005, rel=0.005)

    for _ in range(3):
        save_average_execution_time(one_second)

    assert get_average_execution_time(one_second) == approx(1.0, abs=0.01, rel=0.01)

    for _ in range(2):
        save_average_execution_time(save_average_execution_time, one_second)

    assert get_average_execution_time(one_second) == approx(1.0, abs=0.01, rel=0.01)

    print(get_average_execution_times())

    clear_average_execution_times()

    assert len(get_average_execution_times()) == 0

    for _ in range(3):
        wait_seconds(0.3)

    print(get_average_execution_times())

    assert get_average_execution_time(wait_seconds) == approx(0.3, abs=0.01, rel=0.01)


def test_get_module_location():
    import multiprocessing.dummy

    import egse
    import egse.system
    from egse.system import get_module_location

    # egse is a namespace package which can have different locations!
    assert get_module_location(egse) is None
    assert get_module_location("egse") is None

    # egse.system is a module (.py file) and its location is in the `egse` module/namespace
    assert get_module_location(egse.system) == Path(egse.system.__file__).parent
    assert get_module_location("egse.system") == Path(egse.system.__file__).parent

    # get_module_location is a function, its location is in the system module in the egse namespace
    assert get_module_location(get_module_location) == Path(egse.system.__file__).parent

    # multiprocessing.dummy is a package, so its location is in the multiprocessing folder
    assert get_module_location(multiprocessing.dummy) == Path(multiprocessing.dummy.__file__).parent
    assert get_module_location("multiprocessing.dummy") == Path(multiprocessing.dummy.__file__).parent


def test_get_system_architecture():
    from egse.system import get_system_architecture

    assert "x86" in get_system_architecture() or "arm" in get_system_architecture()


def test_duration():
    dt = datetime.datetime
    td = datetime.timedelta
    tz = datetime.timezone

    now = dt.now(tz=tz.utc)
    dt_start = format_datetime(now)

    dt_end = format_datetime(now + td(seconds=2, milliseconds=500))

    assert duration(dt_start, dt_end).seconds == 2
    assert duration(dt_start, dt_end).total_seconds() == approx(2.5)
    assert duration(dt_start, dt_end) == duration(dt_end, dt_start)

    # The following calls just should raise an exception

    now_plus_half_a_second = now + td(milliseconds=500)
    assert duration(now, now_plus_half_a_second).total_seconds() == approx(0.5)
    assert duration(now_plus_half_a_second, now).total_seconds() == approx(0.5)

    date_1963 = format_datetime(dt(1963, 2, 16, 12, 00, 00, tzinfo=tz.utc))
    date_2023 = format_datetime(dt(2023, 2, 16, 12, 00, 00, tzinfo=tz.utc))
    date_2024 = format_datetime(dt(2024, 2, 16, 12, 00, 00, tzinfo=tz.utc))
    assert str(duration(date_2024, date_1963)) == "22280 days, 0:00:00"
    assert str(duration(date_2024, date_2023)) == "365 days, 0:00:00"

    yesterday = now - td(days=1)
    assert str(duration(now, yesterday)) == "1 day, 0:00:00"
    assert str(duration(now + td(minutes=3.0, seconds=27), yesterday)) == "1 day, 0:03:27"


def test_log_levels_disabled(caplog):
    print()

    import logging

    from egse.system import all_logging_disabled

    with all_logging_disabled(highest_level=logging.WARNING):
        logging.critical("a critical message")
        logging.error("an error message")
        logging.warning("a warning message")
        logging.info("an info message")
        logging.debug("a debug message")

    assert "critical" in caplog.text
    assert "error" in caplog.text
    assert "warning" not in caplog.text
    assert "info" not in caplog.text
    assert "debug" not in caplog.text

    caplog.clear()

    logger = logging.getLogger("testing_logging_disabled")

    with all_logging_disabled(highest_level=logging.INFO):
        logger.critical("a critical message")
        logger.error("an error message")
        logger.warning("a warning message")
        logger.info("an info message")
        logger.debug("a debug message")

        logging.warning("a logging warning message")
        logging.info("a logging info message")

    assert "critical" in caplog.text
    assert "error" in caplog.text
    assert "warning" in caplog.text
    assert "info" not in caplog.text
    assert "debug" not in caplog.text


def test_get_active_loggers():
    from egse.system import get_active_loggers

    assert "testing_active_loggers" not in get_active_loggers()
    assert "testing" not in get_active_loggers()
    assert "testing.active" not in get_active_loggers()
    assert "testing.active.loggers" not in get_active_loggers()

    _ = logging.getLogger("testing_active_loggers")

    assert "testing_active_loggers" in get_active_loggers()

    _ = logging.getLogger("testing.active.loggers")

    assert "testing" in get_active_loggers()
    assert "testing.active" in get_active_loggers()
    assert "testing.active.loggers" in get_active_loggers()


def test_touch():
    touch(__file__)

    assert Path(__file__).exists()

    fn = Path(__file__).parent / "data/counter_test" / "x.cnt"

    try:
        assert not fn.exists()
        touch(fn)
        assert fn.exists()
    finally:
        shutil.rmtree(fn.parent)


def test_is_module():
    import egse.version
    from egse.system import get_os_name

    assert is_module(egse)
    assert is_module("egse")

    assert is_module(egse.version)
    assert is_module("egse.version")

    assert not is_module(get_os_name)  # type: ignore
    assert not is_module("non-existing-module")


def test_is_namespace():
    import egse.version
    from egse.system import get_os_name

    assert is_namespace(egse)
    assert is_namespace("egse")

    assert not is_namespace(egse.version)
    assert not is_namespace("egse.version")

    assert not is_module(get_os_name)  # type: ignore
    assert not is_namespace("non-existing-module")


def test_get_host_ip():
    from egse.system import get_host_ip

    ip = get_host_ip()
    if ip:
        assert "." in ip


def test_do_every(capsys):
    with Timer():
        do_every(0.5, lambda: print(f"{format_datetime()} Hello, World!"), count=0)
    captured = capsys.readouterr()
    assert captured.out == ""

    do_every(0.5, lambda: print(f"{format_datetime()} Hello, World!"), count=1)
    captured = capsys.readouterr()
    assert len(captured.out.rstrip().split("\n")) == 1

    do_every(0.5, lambda: print(f"{format_datetime()} Hello, World!"), count=3)
    captured = capsys.readouterr()
    print("\n".join(captured.out.rstrip().split("\n")))
    assert len(captured.out.rstrip().split("\n")) == 3


def generate_big_data_file(filename: Path, total_n_rows: int = 500_000) -> list[str]:
    """
    Generates a big data file containing 500_000 rows (by default) with each row having a
    timestamp and 10 datapoints of randomized temperatures. The last 10 rows have fixed
    deterministic data that can be tested with pytest.

    Returns:
        The function returns the fixed data rows as a list of str with comma separated values.
            The strings are the same as those written in the CSV file.
    """
    total_rows = 500000
    columns = 10  # Number of temperature columns
    start_date = datetime.datetime(2023, 1, 1, 0, 0, 0)
    time_increment = datetime.timedelta(minutes=1)

    # Deterministic data for the last 10 rows
    deterministic_data = [
        [20.0, 21.5, 22.0, 19.8, 23.1, 22.5, 21.8, 20.5, 19.7, 22.3],
        [20.1, 21.6, 22.1, 19.9, 23.2, 22.6, 21.9, 20.6, 19.8, 22.4],
        [20.2, 21.7, 22.2, 20.0, 23.3, 22.7, 22.0, 20.7, 19.9, 22.5],
        [20.3, 21.8, 22.3, 20.1, 23.4, 22.8, 22.1, 20.8, 20.0, 22.6],
        [20.4, 21.9, 22.4, 20.2, 23.5, 22.9, 22.2, 20.9, 20.1, 22.7],
        [20.5, 22.0, 22.5, 20.3, 23.6, 23.0, 22.3, 21.0, 20.2, 22.8],
        [20.6, 22.1, 22.6, 20.4, 23.7, 23.1, 22.4, 21.1, 20.3, 22.9],
        [20.7, 22.2, 22.7, 20.5, 23.8, 23.2, 22.5, 21.2, 20.4, 23.0],
        [20.8, 22.3, 22.8, 20.6, 23.9, 23.3, 22.6, 21.3, 20.5, 23.1],
        [20.9, 22.4, 22.9, 20.7, 24.0, 23.4, 22.7, 21.4, 20.6, 23.2],
    ]

    def generate_random_temperature_data():
        """Generate random temperature data between 15 and 35 ºC."""
        return [round(random.uniform(15.0, 35.0), 1) for _ in range(columns)]

    def format_timestamp(dt):
        """Format datetime to ISO format."""
        return dt.strftime("%Y-%m-%dT%H:%M:%S")

    print(f"Generating {total_rows} rows of data to {filename}...")

    with open(filename, "w", newline="") as csvfile:
        # Create header: timestamp, temp1, temp2, ..., temp10
        header = ["timestamp"] + [f"temp{i + 1}" for i in range(columns)]

        writer = csv.writer(csvfile)
        writer.writerow(header)

        # Generate random data for most rows
        current_date = start_date
        for i in range(total_rows - 10):
            row_data = [format_timestamp(current_date)] + generate_random_temperature_data()
            writer.writerow(row_data)
            current_date += time_increment

        # Add the last 10 rows with deterministic data
        for i in range(10):
            row_data = [format_timestamp(current_date)] + deterministic_data[i]
            writer.writerow(row_data)
            current_date += time_increment

    return [
        (
            f"{format_timestamp(start_date + time_increment * (total_rows - 10 + i))},"
            f"{','.join([f'{x:3.1f}' for x in row])}"
        )
        for i, row in enumerate(deterministic_data)
    ]


@pytest.mark.asyncio
async def test_periodic():
    count = 0

    def plain_old_function():
        nonlocal count
        logging.info("I'm just a plain old Python function.")
        count += 1
        time.sleep(1.0)

    async def async_function():
        nonlocal count
        logging.info("I'm a brand new async function.")
        count += 1
        await asyncio.sleep(0.5)

    periodic = Periodic(callback=async_function, interval=1.0, skip=False, repeat=5)
    periodic.start()

    # Prevent the test from terminating

    while periodic.is_running():
        try:
            await asyncio.sleep(0.05)
        except asyncio.CancelledError:
            logging.info("periodic_test cancelled")
            logging.info(f"interval ended at {periodic.interval}")
            break

    assert count == 5

    count = 0

    periodic = Periodic(callback=async_function, interval=0.4, skip=True, repeat=5)
    periodic.start()

    # Prevent the test from terminating

    while periodic.is_running():
        try:
            await asyncio.sleep(0.05)
        except asyncio.CancelledError:
            logging.info("periodic_test cancelled")
            logging.info(f"interval ended at {periodic.interval}")
            break

    assert count == 3  # the function call is skipped twice


@pytest.mark.asyncio
async def test_periodic_exception(caplog):
    # So, what happens when a callback throws an exception? Aa error message is logged.

    @execution_count
    def not_good():
        raise ValueError("This is not good!")

    periodic = Periodic(callback=not_good, interval=0.5, skip=True, repeat=2)
    periodic.start()

    await asyncio.sleep(1.2)

    assert "ValueError caught: This is not good!" in caplog.text
    assert not_good.counts() == 2  # type: ignore


def test_camel_to_kebab():
    assert camel_to_kebab("ConfigurationControlServer") == "configuration-control-server"
    assert camel_to_kebab("XMLHttpService") == "xml-http-service"
    assert camel_to_kebab("My123BestProject") == "my123-best-project"

    assert camel_to_kebab("is-already-kebab") == "is-already-kebab"


def test_camel_to_snake():
    assert camel_to_snake("ConfigurationControlServer") == "configuration_control_server"
    assert camel_to_snake("XMLHttpService") == "xml_http_service"
    assert camel_to_snake("My123BestProject") == "my123_best_project"

    assert camel_to_snake("is_already_snake") == "is_already_snake"


def test_recursive_dict_update():
    print()

    top = {"GROUP": {"G1": "initial top-level group"}}
    x = {"NEW_GROUP": {"G2": "additional top-level group"}}

    recursive_dict_update(top, x)

    assert "NEW_GROUP" in top
    assert "initial" in top["GROUP"]["G1"]
    assert "additional" in top["NEW_GROUP"]["G2"]

    x = {"NEW_SUBGROUP": "a new subgroup added to NEW_GROUP"}

    recursive_dict_update(top["NEW_GROUP"], x)

    rich.print(top)

    assert "NEW_SUBGROUP" in top["NEW_GROUP"]

    x = {"GROUP": {"G3": "a new group added to top-level GROUP"}}

    recursive_dict_update(top, x)

    rich.print(top)

    assert "G3" in top["GROUP"]


def test_redirect_output_to_log():
    import egse.env

    # The redirected log file will be created in the current folder for this test

    with patch.object(egse.env, "get_log_file_location", return_value=Path(HERE)):
        log = redirect_output_to_log("test_redirect_output_to_log.log")
        log.write("This is a test message to be logged.\n")
        log.close()

        log_path = Path(HERE) / "test_redirect_output_to_log.log"

        assert log_path.exists()

        with open(log_path, "r") as f:
            content = f.read()
            assert "This is a test message to be logged." in content

        # Test appending to the log file

        log = redirect_output_to_log("test_redirect_output_to_log.log", append=True)
        log.write("Appending a new test message to be logged.\n")
        log.close()

        with open(log_path, "r") as f:
            content = f.read()
            assert "Appending a new test message to be logged." in content

        log = redirect_output_to_log("test_redirect_output_to_log.log", append=False, overwrite=True)
        log.write("Overwriting a new test message to be logged.\n")
        log.close()

        with open(log_path, "r") as f:
            content = f.read()
            assert content.splitlines() == ["Overwriting a new test message to be logged."]

        # Test overwriting the log file.
        # The overwrite argument is False, so it should raise an exception

        with pytest.raises(FileExistsError):
            log = redirect_output_to_log("test_redirect_output_to_log.log", overwrite=False)

        log_path.unlink()  # Clean up the log file after test

        # Test the behavior when the output_fn is absolute and the directory does not exist.

        absolute_log_path = Path(HERE) / "xxx1234/absolute_test_log.log"

        log = redirect_output_to_log(str(absolute_log_path))
        log.write("This is a test message to be logged in an absolute path.\n")
        log.close()

        assert absolute_log_path.exists()

        with open(absolute_log_path, "r") as f:
            content = f.read()
            assert "This is a test message to be logged in an absolute path." in content

        absolute_log_path.unlink()  # Clean up the absolute log file after test
        absolute_log_path.parent.rmdir()  # Remove the created directory



"""Device connection interface for the Ariel Telescope Control Unit (TCU)."""

import logging
import time

import serial

from egse.device import DeviceConnectionInterface, DeviceTransport
from egse.settings import Settings

logger = logging.getLogger(__name__)
DEVICE_SETTINGS = Settings.load("Ariel TCU Controller")

EXPECTED_TRANSACTION_ID = -1


class TcuError(Exception):
    """Generic TCU error for low-level classes."""

    pass


class TcuDeviceInterface(DeviceConnectionInterface, DeviceTransport):
    """Defines the low-level interface to the Ariel TCU (Arduino)."""

    def __init__(self, port: int = None):
        """Initialisation of a serial interface to the TCU Arduino.

        Args:
            port (int): Serial port to which to connect to the TCU Arduino.
        """

        super().__init__()

        # self.hostname = hostname or DEVICE_SETTINGS["HOSTNAME"]
        self.port = port or DEVICE_SETTINGS["COM_PORT"]

        self.arduino = serial.Serial()
        self.arduino.port = self.port
        self.arduino.baudrate = DEVICE_SETTINGS["BAUD_RATE"]
        self.arduino.bytesize = DEVICE_SETTINGS["NUM_DATA_BITS"]
        self.arduino.parity = DEVICE_SETTINGS["PARITY"]
        self.arduino.stopbits = DEVICE_SETTINGS["NUM_STOP_BITS"]

    def connect(self) -> None:
        """Opens the serial port to the TCU Arduino.

        Raises:
            TcuError: When the serial port could not be opened.
        """

        if self.is_connected():
            raise TcuError("TCU already connected.")
        # if self.hostname in (None, ""):
        #     raise TcuError("TCU hostname is not initialised.")
        if self.port in (None, 0):
            raise TcuError("TCU serial port is not initialised.")

        try:
            self.arduino.open()
        except Exception as exc:
            raise TcuError(f"Failed to open TCU serial port {self.port}") from exc

    def disconnect(self) -> None:
        """Closes the serial port to the TCU Arduino.

        Raises:
            TcuError: When the serial port could not be closed.
        """

        try:
            self.arduino.close()
        except Exception as exc:
            raise TcuError(f"Failed to close TCU serial port {self.port}") from exc

    def reconnect(self) -> None:
        """Re-connects to the Ariel TCU Arduino."""

        if self.is_connected():
            self.disconnect()
        self.connect()

    def is_connected(self) -> bool:
        """Checks whether the serial port to the TCU Arduino is open.

        Returns:
            True if the serial port to the TCU Arduino is open; False otherwise.
        """

        return self.arduino.is_open

    def trans(self, command: str) -> bytes:
        """Sends the given command to the TCU Arduino and returns the response.

        This is seen as a transaction.

        Args:
            command: Command string to send to the TCU Arduino.

        Returns:
            Response string from the TCU Arduino.
        """

        self.arduino.write(command.encode("utf-8"))
        time.sleep(0.05)
        return self.arduino.readline()


class TcuHexInterface(DeviceTransport):
    """Transport class to generate the hex strings that are sent to the TCU."""

    def trans(self, command: str) -> bytes:
        """Returns the given command string as bytes.

        This can be used to:

            - Test whether the command string is built correctly (and the transaction identifier is incremented for
              each command call);
            - Generate the hex string that is sent to the TCU, such that it can be displayed in the TCU UI.

        Args:
            command (str): Command string to encode.

        Returns:
            Encoded command string.
        """

        global EXPECTED_TRANSACTION_ID
        EXPECTED_TRANSACTION_ID += 1

        return command.encode()
